import lusid


def connect(config):
    lusid_config = lusid.Configuration()
    lusid_config.access_token = config["token"]
    lusid_config.host = config["apiUrl"]

    return lusid.SyncApiClient(lusid_config), lusid



import os
import unittest
from parameterized import parameterized
import lusid
from pathlib import Path
import finbourne_sdk_utils.cocoon as cocoon
import pandas as pd
import numpy as np
from finbourne_sdk_utils import logger
from .mock_api_factory import MockApiFactory


class CocoonPropertiesTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Use a mock of the lusid.ApiClientFactory
        cls.api_factory = MockApiFactory( )
        cls.logger = logger.LusidLogger(os.getenv("FBN_LOG_LEVEL", "info"))

    @parameterized.expand(
        [
            ["Instrument/default/Figi", [True, "string"]],
            ["Instrument/default/PropertyThatDoesNotExist", [False, None]],
            ["Transaction/default/TradeToPortfolioRate", [True, "number"]],
            ["Transaction/Operations/Strategy", [True, "string"]],
            ["Holding/Operations/Currency", [True, "currency"]],
            ["Instrument/default/Forbidden", [False, None]],
        ]
    )
    def test_check_property_definitions_exist_in_scope_single(
        self, property_key, expected_outcomes, throws_exception=False
    ) -> None:
        """
        Tests that checking for a property definition in a single scope works as expected. The call to LUSID
        is mocked here

        :param str property_key: The property key to check for
        :param list[bool, str] expected_outcomes: Whether or not the property is expected to exist and if it does its type

        :return: None
        """
        try:
            (
                property_existence,
                property_type,
            ) = cocoon.properties.check_property_definitions_exist_in_scope_single(
                api_factory=self.api_factory, property_key=property_key
            )

            self.assertEqual(property_existence, expected_outcomes[0])
            self.assertEqual(property_type, expected_outcomes[1])
        except lusid.exceptions.ApiException:
            self.assertEqual(throws_exception, True)

    @parameterized.expand(
        [
            # Check for two properties which don't exist and one that does with a type of string
            [
                "default",
                "Instrument",
                pd.DataFrame(
                    data={
                        "Figi": ["BBG009QYHGN8", "BBG00JRKP9P0", "BBG009QYHGN8"],
                        "MoodysCreditRating": ["A2", "A1", "A2"],
                        "Age": [30, 70, 15],
                    },
                ),
                ["Figi", "MoodysCreditRating", "Age"],
                [
                    ["MoodysCreditRating", "Age"],
                    pd.DataFrame(
                        data={
                            "Figi": ["BBG009QYHGN8", "BBG00JRKP9P0", "BBG009QYHGN8"],
                            "MoodysCreditRating": ["A2", "A1", "A2"],
                            "Age": [30, 70, 15],
                        },
                    ),
                ],
            ],
            # Check for one property that exists with a currency type and one that does not exist
            [
                "Operations",
                "Holding",
                pd.DataFrame(
                    data={
                        "Currency": ["GBP", "USD", "AUD"],
                        "Notional": [True, False, True],
                    }
                ),
                ["Currency", "Notional"],
                [
                    ["Notional"],
                    pd.DataFrame(
                        data={
                            "Currency": ["GBP", "USD", "AUD"],
                            "Notional": [True, False, True],
                        }
                    ),
                ],
            ],
        ]
    )
    def test_check_property_definitions_exist_in_scope(
        self, scope, domain, data_frame, property_columns, expected_outcome
    ) -> None:
        """
        Tests that checking of the existence of a set of property definitions in a scope works as expected

        :param str scope: The scope of the properties to check for
        :param str domain: The domain of the properties to check for
        :param pd.DataFrame data_frame: The dataframe containing the column headers and data to use to determine data types
        :param list[str]: The list of property columns to check for
        :param list[list[str], pd.DataFrame] expected_outcome: The expected result
        :return: None
        """

        (
            missing_columns,
            updated_data_frame,
        ) = cocoon.properties.check_property_definitions_exist_in_scope(
            api_factory=self.api_factory,
            domain=domain,
            data_frame=data_frame,
            target_columns=property_columns,
            column_to_scope={column: scope for column in property_columns},
        )

        self.assertEqual(first=len(missing_columns), second=len(set(missing_columns)))

        self.assertEqual(first=set(missing_columns), second=set(expected_outcome[0]))

        self.assertTrue(expr=updated_data_frame.equals(expected_outcome[1]))

    @parameterized.expand(
        [
            # Test that a missing property is created as expected
            [
                "CreditRating",
                "Instrument",
                pd.DataFrame(
                    data=[
                        {
                            "instrument_name": "GlobalCreditFund",
                            "lookthrough_scope": "SingaporeBranch",
                            "lookthrough_code": "PORT_12490FKS9",
                            "format": "json",
                            "contract": "{asset_type: 'fund', term_sheet: 'GlobalCreditFundConstituents.pdf'}",
                            "Moodys": "A2",
                        }
                    ]
                ),
                ["Moodys"],
                [
                    {"Moodys": "Instrument/CreditRating/Moodys"},
                    pd.DataFrame(
                        data=[
                            {
                                "instrument_name": "GlobalCreditFund",
                                "lookthrough_scope": "SingaporeBranch",
                                "lookthrough_code": "PORT_12490FKS9",
                                "format": "json",
                                "contract": "{asset_type: 'fund', term_sheet: 'GlobalCreditFundConstituents.pdf'}",
                                "Moodys": "A2",
                            }
                        ]
                    ),
                ],
            ],
            # Test that a property with all null values has a definition created as a string not a number
            [
                "CreditRating",
                "Instrument",
                pd.DataFrame(
                    data=[
                        {
                            "instrument_name": "GlobalCreditFund",
                            "lookthrough_scope": "SingaporeBranch",
                            "lookthrough_code": "PORT_12490FKS9",
                            "format": "json",
                            "contract": "{asset_type: 'fund', term_sheet: 'GlobalCreditFundConstituents.pdf'}",
                            "S&P": np.nan,
                        }
                    ]
                ),
                ["S&P"],
                [
                    {"S&P": "Instrument/CreditRating/SandP"},
                    pd.DataFrame(
                        data=[
                            {
                                "instrument_name": "GlobalCreditFund",
                                "lookthrough_scope": "SingaporeBranch",
                                "lookthrough_code": "PORT_12490FKS9",
                                "format": "json",
                                "contract": "{asset_type: 'fund', term_sheet: 'GlobalCreditFundConstituents.pdf'}",
                                "S&P": np.nan,
                            }
                        ],
                        dtype=object,
                    ),
                ],
            ],
        ]
    )
    def test_create_property_definitions_from_file(
        self, scope, domain, data_frame, missing_property_columns, expected_outcome
    ) -> None:
        """

        :param str scope: The scope to create the property definitions in
        :param str domain: The domain to create the property definitions in
        :param pd.Series data_frame_dtypes: The dataframe dtypes to add definitions for
        :return: dict property_key_mapping: A mapping of data_frame columns to property keys
        :param [dict, pd.DataFrame] expected_outcome: The expected outcome

        :return: None
        """

        (
            property_key_mapping,
            data_frame_updated,
        ) = cocoon.properties.create_property_definitions_from_file(
            api_factory=self.api_factory,
            domain=domain,
            data_frame=data_frame,
            missing_property_columns=missing_property_columns,
            column_to_scope={column: scope for column in missing_property_columns},
        )

        self.assertEqual(first=property_key_mapping, second=expected_outcome[0])

        self.assertTrue(expr=data_frame_updated.equals(expected_outcome[1]))

    @parameterized.expand(
        [
            # Test creating property values in the instrument domain
            [
                pd.Series(data=["A2", "A-"], index=["Moodys", "S&P"]),
                "CreditRating",
                "Instrument",
                pd.Series(data=["object", "object"], index=["Moodys", "S&P"]),
                [
                    lusid.models.PerpetualProperty(
                        key="Instrument/CreditRating/Moodys",
                        value=lusid.models.PropertyValue(label_value="A2"),
                    ),
                    lusid.models.PerpetualProperty(
                        key="Instrument/CreditRating/SandP",
                        value=lusid.models.PropertyValue(label_value="A-"),
                    ),
                ],
            ],
            # Test creating property values in the transaction domain, including a metric value property
            [
                pd.Series(
                    data=["TD_1241247", 30], index=["TraderId", "Rebalancing_Interval"]
                ),
                "Operations",
                "Transaction",
                pd.Series(
                    data=["object", "float64"],
                    index=["TraderId", "Rebalancing_Interval"],
                ),
                {
                    "Transaction/Operations/TraderId": lusid.models.PerpetualProperty(
                        key="Transaction/Operations/TraderId",
                        value=lusid.models.PropertyValue(label_value="TD_1241247"),
                    ),
                    "Transaction/Operations/Rebalancing_Interval": lusid.models.PerpetualProperty(
                        key="Transaction/Operations/Rebalancing_Interval",
                        value=lusid.models.PropertyValue(
                            metric_value=lusid.models.MetricValue(value=30)
                        ),
                    ),
                },
            ],
        ]
    )
    def test_create_property_values(
        self, row, scope, domain, dtypes, expected_outcome
    ) -> None:
        """

        :param pd.Series row:
        :param str scope:
        :param str domain:
        :param pd.Series dtypes:
        :param list[lusid.models.PerpetualProperty] expected_outcome:

        :return: None
        """

        property_values = cocoon.properties.create_property_values(
            row=row, column_to_scope={}, scope=scope, domain=domain, dtypes=dtypes
        )

        self.assertEqual(first=property_values, second=expected_outcome)

    @parameterized.expand(
        [
            [
                "Just a code",
                ["Strategy"],
                "Traders",
                "Transaction",
                ["Transaction/Traders/Strategy"],
            ],
            [
                "A scope and code",
                ["Signal/Strategy"],
                "Traders",
                "Transaction",
                ["Transaction/Signal/Strategy"],
            ],
            [
                "A full key",
                ["Portfolio/Signal/Strategy"],
                "Traders",
                "Transaction",
                ["Portfolio/Signal/Strategy"],
            ],
            [
                "A scope which is the same as the domain",
                ["Transaction/Strategy"],
                "Traders",
                "Transaction",
                ["Transaction/Transaction/Strategy"],
            ],
        ]
    )
    def test_infer_full_property_keys(
        self, _, partial_keys, properties_scope, domain, expected_outcome
    ):

        full_keys = cocoon.properties._infer_full_property_keys(
            partial_keys=partial_keys, properties_scope=properties_scope, domain=domain
        )

        self.assertEqual(
            first=expected_outcome,
            second=full_keys,
            msg="The full keys don't matched the expected outcome",
        )
    def test_create_property_definitions_with_leading_nulls(self):
        """
        Tests that property definitions are created with correct data types
        even when columns have leading null values.
        
        The function should find the first non-null value and infer the type
        from that, rather than defaulting to string for columns with null headers.
        """
        # Create a DataFrame with leading nulls in both string and numeric columns
        data_frame = pd.DataFrame(
            data=[
                {
                    "Name": np.nan,
                    "Count": np.nan,
                    "Rating": np.nan,
                },
                {
                    "Name": np.nan,
                    "Count": np.nan,
                    "Rating": np.nan,
                },
                {
                    "Name": "Product A",  # First non-null string value
                    "Count": 42,  # First non-null integer value
                    "Rating": 4.5,  # First non-null float value
                },
                {
                    "Name": "Product B",
                    "Count": 100,
                    "Rating": 3.8,
                },
            ]
        )

        missing_property_columns = ["Name", "Count", "Rating"]
        column_to_scope = {col: "default" for col in missing_property_columns}

        # Call the function
        property_key_mapping, updated_data_frame = cocoon.properties.create_property_definitions_from_file(
            api_factory=self.api_factory,
            domain="Instrument",
            data_frame=data_frame,
            missing_property_columns=missing_property_columns,
            column_to_scope=column_to_scope,
        )

        # Validate that properties were created
        self.assertEqual(len(property_key_mapping), 3)
        self.assertIn("Name", property_key_mapping)
        self.assertIn("Count", property_key_mapping)
        self.assertIn("Rating", property_key_mapping)

        # Validate property keys were created correctly
        self.assertEqual(
            property_key_mapping["Name"],
            "Instrument/default/Name"
        )
        self.assertEqual(
            property_key_mapping["Count"],
            "Instrument/default/Count"
        )
        self.assertEqual(
            property_key_mapping["Rating"],
            "Instrument/default/Rating"
        )

        # Validate that the DataFrame data types were inferred correctly
        # Count should be int64 or float64 (numeric)
        self.assertIn(str(updated_data_frame["Count"].dtype), ["int64", "float64"])
        # Rating should be float64 (numeric)
        self.assertEqual(str(updated_data_frame["Rating"].dtype), "float64")
        # Name should remain object (string)
        self.assertEqual(str(updated_data_frame["Name"].dtype), "object")


import pandas as pd

from finbourne_sdk_utils.lpt import lpt
from finbourne_sdk_utils.lpt import lse
from finbourne_sdk_utils.lpt import stdargs

TOOLNAME = "instr"
TOOLTIP = "Display specified Instruments"


def parse(extend=None, args=None):
    return (
        stdargs.Parser("Get Instruments", ["filename", "limit", "asat"])
        .add("instrument", nargs="*")
        .add("--type", "-t", default="ClientInternal", help="Instrument type")
        .add("--identifiers", nargs="*", help="identifiers to display")
        .add("--properties", nargs="*", help="properties required for display")
        .add(
            "--from",
            dest="from_file",
            nargs=2,
            metavar=("filename.csv", "column"),
            help="load values from file",
        )
        .add("--effective_at")
        .extend(extend)
        .parse(args)
    )


def process_args(api, args):
    MAX_PROPS = 50

    def success(result):
        idents = [f"identifiers.KEY:{v}" for v in args.identifiers]
        columns = ["lusid_instrument_id", "name"]
        columns.extend(idents)
        if args.properties:
            columns.extend(["P:" + v for v in args.properties])

        df = lpt.to_df(result.content.values.values(), columns)
        return df.rename(columns=dict(zip(idents, args.identifiers)))

    if args.from_file:
        df = pd.read_csv(args.from_file[0])[args.from_file[1:]].drop_duplicates()
        args.instrument.extend(df[args.from_file[1]].values)

    def step3(result, props_remaining, final_dataframe):
        df = success(result)

        if final_dataframe is None:
            final_dataframe = df
        else:
            props = [c for c in df.columns.values if c.startswith("P:")]
            final_dataframe = final_dataframe.merge(
                df[["lusid_instrument_id"] + props],
                how="left",
                on="lusid_instrument_id",
            )

        if len(props_remaining) > 0:
            args.properties = props_remaining[:MAX_PROPS]
            remaining = props_remaining[MAX_PROPS:]
            return api.call.get_instruments(
                args.type, request_body=args.instrument, property_keys=args.properties
            ).bind(lambda r: step3(r, remaining, final_dataframe))
        return final_dataframe

    def step2(result=None):
        next_step = success
        if result is not None:
            # Contains full set of instrument properties
            l = list(
                result.apply(lambda r: f"Instrument/{r['scope']}/{r['code']}", axis=1)
            )
            args.properties = l[:MAX_PROPS]  # limitation
            if len(l) > MAX_PROPS:
                next_step = lambda r: step3(r, l[MAX_PROPS:], None)

        return api.call.get_instruments(
            args.type,
            request_body=args.instrument,
            property_keys=args.properties,
            effective_at=lpt.to_date(args.effective_at),
        ).bind(next_step)

    if args.identifiers is None:
        args.identifiers = [args.type]

    if args.properties:
        if args.properties[0] == "all":
            from finbourne_sdk_utils.lpt import qry_properties as qp

            return qp.process_args(api, qp.parse(args=["-d", "Instrument"])).bind(step2)

    return step2()


# Standalone tool
def main(parse=parse, display_df=lpt.display_df):
    return lpt.standard_flow(parse, lse.connect, process_args, display_df)



import os
from urllib.request import quote

import lusid

from .refreshing_token import RefreshingToken

config_mapping = {
    "FBN_TOKEN_URL": "tokenUrl",
    "FBN_USERNAME": "username",
    "FBN_PASSWORD": "password",
    "FBN_CLIENT_ID": "clientId",
    "FBN_CLIENT_SECRET": "clientSecret",
    "FBN_LUSID_API_URL": "apiUrl",
}


def check_for_missing_config(config):
    return [
        {"Env variable": env_var, "Secrets file key": config_key}
        for env_var, config_key in config_mapping.items()
        if not os.getenv(env_var, config["api"][config_key])
    ]


def connect(config, **kwargs):
    if "api" not in config.keys():
        config["api"] = {}
        config["api"]["tokenUrl"] = None
        config["api"]["username"] = None
        config["api"]["password"] = None
        config["api"]["clientId"] = None
        config["api"]["clientSecret"] = None
        config["api"]["apiUrl"] = None

    missing_config = check_for_missing_config(config)
    if len(missing_config) > 0:
        raise Exception(f"Missing the following config: {missing_config}")

    token_url = os.getenv("FBN_TOKEN_URL", config["api"]["tokenUrl"])
    username = os.getenv("FBN_USERNAME", config["api"]["username"])
    password = quote(os.getenv("FBN_PASSWORD", config["api"]["password"]), "*!")
    client_id = quote(os.getenv("FBN_CLIENT_ID", config["api"]["clientId"]), "*!")
    client_secret = quote(
        os.getenv("FBN_CLIENT_SECRET", config["api"]["clientSecret"]), "*!"
    )
    api_url = os.getenv("FBN_LUSID_API_URL", config["api"]["apiUrl"])

    token_request_body = (
        "grant_type=password&username={0}".format(username)
        + "&password={0}&scope=openid client groups".format(password)
        + "&client_id={0}&client_secret={1}".format(client_id, client_secret)
    )

    headers = {
        "Accept": "application/json",
        "Content-Type": "application/x-www-form-urlencoded",
    }

    config = lusid.Configuration()
    config.access_token = RefreshingToken(token_url, token_request_body, headers)
    config.host = api_url

    return (lusid.SyncApiClient(configuration=config), lusid)



import os

import lusid
import lusid.models as models
import unittest


from finbourne_sdk_utils import logger
from finbourne_sdk_utils.cocoon.cocoon_printer import (
    format_portfolios_response,
    format_instruments_response,
    format_holdings_response,
    format_quotes_response,
    format_transactions_response,
    get_portfolio_from_href,
    format_reference_portfolios_response,
    CocoonPrinter,
)
from parameterized import parameterized

instrument_id = "LUID_01234567"


instrument_success = models.UpsertInstrumentsResponse(
    values={
        "LUID_01234567": models.Instrument(
            lusid_instrument_id="LUID_01234567",
            version= models.Version(as_at_version_number=1,effective_from="2010-09-01T09:31:22.664000+00:00",asAtDate="2019-09-01T09:31:22.664000+00:00"),
            name="name1",
            identifiers={
                "Luid": "LUID_01234567",
                "Figi": "BBG000BLNNH6"
            },
            state="Active",
        )
    },
    failed={
        "error101": models.ErrorDetail(
            id="LUID_01234567",
            type="error",
            detail="failed at line 42 in file root.trunk.branch.py",
        )
    },
)

portfolio_success = models.Portfolio(
    links=[],
    version= models.Version(as_at_version_number=1,effective_from="2010-09-01T09:31:22.664000+00:00",asAtDate="2019-09-01T09:31:22.664000+00:00"),
    type="Transaction",
    display_name="name2",
    description="test portfolio",
    created="2010-09-01T09:31:22.664000+00:00",
    href="https://www.tes.lusid/code/portfolio?asAt=2019-12-05T10%3A25%3A45.6141270%2B00%3A00",
    id=models.ResourceId(code="ID00001", scope="default test"),
)

transaction_success = models.UpsertPortfolioTransactionsResponse(
    href="https://www.notadonaim.lusid.com/api/api/code/transactions?asAt=2019-12-05T10%3A25%3A45.6141270%2B00%3A00",
    links=[],
    version= models.Version(as_at_version_number=1,effective_from="2010-09-01T09:31:22.664000+00:00",asAtDate="2019-09-01T09:31:22.664000+00:00"),
)

quote_success = models.UpsertQuotesResponse(
    failed={
        "BBG001MM1KV4-Figi_2019-10-28": models.ErrorDetail(
            id="BBG001MM1KV4",
            type="Price",
            details="Quote had an error",
        )
    },
    values={
        "BBG001MM1KV4-Figi_2019-10-28": models.Quote(
            as_at="2019-01-16T09:31:22.664000+00:00",
            quote_id=models.QuoteId(
                quote_series_id=models.QuoteSeriesId(
                    provider="Default",
                    instrument_id="BBG001MM1KV4",
                    instrument_id_type="Figi",
                    quote_type="Price",
                    field="Mid",
                ),
                effective_at="2019-01-16T09:31:22.664000+00:00",
            ),
            uploaded_by="test",
        )
    },
)

adjust_holding_success = models.AdjustHolding(
    href="https://notadomain.lusid.com/api/api/code/holdings?asAt=2019-12-05T10%3A25%3A45.6141270%2B00%3A00",
    version=models.Version(as_at_version_number=1,effective_from="2010-09-01T09:31:22.664000+00:00",asAtDate="2019-09-01T09:31:22.664000+00:00"),
)

reference_portfolio_success = models.Portfolio(
    links=[],
    version=models.Version(as_at_version_number=1,effective_from="2010-09-01T09:31:22.664000+00:00",asAtDate="2019-09-01T09:31:22.664000+00:00"),
    type="Transaction",
    display_name="name3",
    description="test reference portfolio",
    created="2016-01-01T09:31:22.664000+00:00",
    href="https://www.tes.lusid/code/portfolio?asAt=2019-12-05T10%3A25%3A45.6141270%2B00%3A00",
    id=models.ResourceId(code="ID00002", scope="default test"),
)


class MockHttpResp:
    status = "404"
    reason = "not found"
    data = b'{"name":"TestFailure","errorDetails":[],"code":404,"type":"https://docs.lusid.com/#section/Error-Codes/404","title":"This is a test error title","status":404,"detail":"Test detail","instance":"https://test.lusid.com/app/insights/logs/TestRequestID-0001","extensions":{}}'

    @classmethod
    def getheaders(cls):
        return {"lusid-meta-requestId": "TestRequestID-0001"}


api_exception = lusid.exceptions.ApiException(http_resp=MockHttpResp())


# build finbourne_sdk_utils responses

responses = {
    "instruments": {
        "errors": [api_exception for _ in range(2)],
        "success": [instrument_success for _ in range(2)],
    },
    "portfolios": {
        "errors": [api_exception for _ in range(2)],
        "success": [portfolio_success for _ in range(2)],
    },
    "transactions": {
        "errors": [api_exception for _ in range(2)],
        "success": [transaction_success for _ in range(2)],
    },
    "quotes": {
        "errors": [api_exception for _ in range(2)],
        "success": [quote_success for _ in range(2)],
    },
    "holdings": {
        "errors": [api_exception for _ in range(2)],
        "success": [adjust_holding_success for _ in range(2)],
    },
    "reference_portfolios": {
        "errors": [api_exception for _ in range(2)],
        "success": [reference_portfolio_success for _ in range(2)],
    },
}

empty_response_with_full_shape = {
    "instruments": {
        "errors": [],
        "success": [
            models.UpsertInstrumentsResponse(values={}, failed={}),
        ],
    },
    "portfolios": {
        "errors": [],
        "success": [],
    },
    "transactions": {
        "errors": [],
        "success": [],
    },
    "quotes": {
        "errors": [],
        "success": [models.UpsertQuotesResponse(failed={}, values={})],
    },
    "holdings": {
        "errors": [],
        "success": [],
    },
    "reference_portfolios": {
        "errors": [],
        "success": [],
    },
}

empty_response_missing_shape = {
    "instruments": {
        "errors": [],
        "success": [],
    },
    "portfolios": {
        "errors": [],
        "success": [],
    },
    "transactions": {
        "errors": [],
        "success": [],
    },
    "quotes": {"errors": [], "success": []},
    "holdings": {
        "errors": [],
        "success": [],
    },
    "reference_portfolios": {
        "errors": [],
        "success": [],
    },
}

responses_no_error_field = {
    "instruments": {
        "success": [instrument_success],
    },
    "portfolios": {
        "success": [portfolio_success],
    },
    "transactions": {
        "success": [transaction_success],
    },
    "quotes": {"success": [quote_success]},
    "holdings": {
        "success": [adjust_holding_success],
    },
    "reference_portfolios": {
        "success": [portfolio_success],
    },
}
responses_no_success_field = {
    "instruments": {
        "errors": [api_exception for _ in range(2)],
    },
    "portfolios": {
        "errors": [api_exception for _ in range(2)],
    },
    "transactions": {
        "errors": [api_exception for _ in range(2)],
    },
    "quotes": {
        "errors": [api_exception for _ in range(2)],
    },
    "holdings": {
        "errors": [api_exception for _ in range(2)],
    },
    "reference_portfolios": {
        "errors": [api_exception for _ in range(2)],
    },
}

extended_error_expected = [
    ["not found", "TestRequestID-0001"],
    ["not found", "TestRequestID-0001"],
]


class CocoonPrinterTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.logger = logger.LusidLogger(os.getenv("FBN_LOG_LEVEL", "info"))

    def assert_responses(
        self,
        num_items,
        expected_value,
        succ=None,
        err=None,
        failed=None,
        err_extended=False,
        data_entity_details=False,
    ):
        if succ is not None and not data_entity_details:
            self.assertEqual(num_items, len(succ))
            for index, row in succ.iterrows():
                self.assertEqual(expected_value["succ"][index], row[succ.columns[0]])

        # removed datetime from tests as the json flatten has unexpected behaviour for datetime, seems to use a Timestamp(str)
        # datetimes not incluced in the V1 test coverage
        # also appears to use the alias varient of the property 
        if succ is not None and data_entity_details:
            self.assertEqual(num_items, len(succ))
            for index, row in succ.iterrows():
                self.assertEqual(
                    expected_value["succ"][index],
                    {k: v for (k, v) in row.to_dict().items() if v and ("asAtDate" in k) == False and ("effectiveFrom" in k )== False }, 
                )

        if err is not None and not err_extended:
            self.assertEqual(num_items, len(err))
            for index, row in err.iterrows():
                self.assertEqual(expected_value["err"][index], row[err.columns[0]])

        if err is not None and err_extended:
            self.assertEqual(num_items, len(err))
            for index, row in err.iterrows():
                self.assertEqual(expected_value["err"][index][0], row[err.columns[0]])
                self.assertEqual(expected_value["err"][index][1], row[err.columns[2]])

        if failed is not None and not data_entity_details:
            self.assertEqual(num_items, len(failed))
            for index, row in failed.iterrows():
                self.assertEqual(
                    expected_value["failed"][index], row[failed.columns[0]]
                )

        if failed is not None and data_entity_details:
            self.assertEqual(num_items, len(failed))
            for index, row in failed.iterrows():
                self.assertEqual(
                    expected_value["failed"][index],
                    {k: v for (k, v) in row.to_dict().items() if v},
                )

    @parameterized.expand(
        [
            (
                "standard_quotes",
                [
                    "asdfasdf/FDaFA/ FFDSSD/234^$^&SEF/code/quotes?asAt=2019-12-05T10%3A25%3A45.6141270%2B00%3A00",
                    "asdfasdf/FDaFA/ FFDSSD/234^$^&SEF/code1/quotes?asAt=2019-12-05T10%3A25%3A45.6141270%2B00%3A00",
                    "asdfasdf/FDaFA/ FFDSSD/234^$^&SEF/code12/quotes?asAt=2019-12-05T10%3A25%3A45.6141270%2B00%3A00",
                ],
                "quotes",
                ["code", "code1", "code12"],
            ),
            (
                "standard_holdings",
                [
                    "asdfasdf/FDaFA/ FFDSSD/234^$^&SEF/code/holdings?asAt=2019-12-05T10%3A25%3A45.6141270%2B00%3A00",
                    "asdfasdf/FDaFA/ FFDSSD/234^$^&SEF/code1/holdings?asAt=2019-12-05T10%3A25%3A45.6141270%2B00%3A00",
                    "asdfasdf/FDaFA/ FFDSSD/234^$^&SEF/code12/holdings?asAt=2019-12-05T10%3A25%3A45.6141270%2B00%3A00",
                ],
                "holdings",
                ["code", "code1", "code12"],
            ),
        ]
    )
    def test_get_portfolio_from_href(self, _, href, file_type, expected_value):
        codes = get_portfolio_from_href(href=href, file_type=file_type)
        self.assertEqual(len(href), len(codes))
        self.assertEqual(expected_value, codes)

    @parameterized.expand(
        [
            (
                "standard_response",
                responses,
                2,
                {
                    "succ": [
                        instrument_id,
                        instrument_id,
                    ],
                    "failed": [
                        "error101",
                        "error101",
                    ],
                    "err": ["not found", "not found"],
                },
                False,
                False,
            ),
            (
                "standard_response_fully_expanded",
                responses,
                2,
                {
                    "succ": [
                        {
                            "lusidInstrumentId": "LUID_01234567",
                            "version.asAtVersionNumber":1,
                            "state": "Active",
                            "identifiers.Luid": "LUID_01234567",
                            "identifiers.Figi": "BBG000BLNNH6",
                            "name": "name1",
                        },
                        {
                            "lusidInstrumentId": "LUID_01234567",
                            "version.asAtVersionNumber":1,
                            "state": "Active",
                            "identifiers.Luid": "LUID_01234567",
                            "identifiers.Figi": "BBG000BLNNH6",
                            "name": "name1",
                        },
                    ],
                    "failed": [
                        {
                            "id": "LUID_01234567",
                            "detail": "failed at line 42 in file root.trunk.branch.py",
                            "type": "error",
                        },
                         {
                            "id": "LUID_01234567",
                            "detail": "failed at line 42 in file root.trunk.branch.py",
                            "type": "error",
                        },
                    ],
                    "err": ["not found", "not found"],
                },
                False,
                True,
            ),
            ("empty_response", empty_response_with_full_shape, 0, {}, False, False),
            (
                "empty_response_missing_shape",
                empty_response_missing_shape,
                0,
                {},
                False,
                False,
            ),
            (
                "standard_response_with_extended_errors",
                responses,
                2,
                {
                    "succ": [
                        "LUID_01234567",
                        "LUID_01234567",
                    ],
                    "failed": [
                        "error101",
                        "error101",
                    ],
                    "err": extended_error_expected,
                },
                True,
                False,
            ),
        ]
    )
    def test_format_instruments_response_success(
        self,
        _,
        response,
        num_items,
        expected_value,
        extended_errors,
        data_entity_details,
    ):
        succ, err, failed = format_instruments_response(
            response,
            extended_error_details=extended_errors,
            data_entity_details=data_entity_details,
        )
        self.assert_responses(
            num_items,
            expected_value,
            succ=succ,
            err=err,
            failed=failed,
            err_extended=extended_errors,
            data_entity_details=data_entity_details,
        )

        printer = self.create_printer(
            response,
            "instruments",
            extended_errors=extended_errors,
            data_entity_details=data_entity_details,
        )
        succ, err, failed = printer.format_response()
        self.assert_responses(
            num_items,
            expected_value,
            succ=succ,
            err=err,
            failed=failed,
            err_extended=extended_errors,
            data_entity_details=data_entity_details,
        )

    @parameterized.expand(
        [
            (
                "standard_response",
                responses,
                2,
                {"succ": ["ID00001", "ID00001"], "err": ["not found", "not found"]},
                False,
            ),
            ("empty_response", empty_response_with_full_shape, 0, {}, False),
            (
                "standard_response_with_extended_errors",
                responses,
                2,
                {"succ": ["ID00001", "ID00001"], "err": extended_error_expected},
                True,
            ),
        ]
    )
    def test_format_portfolios_response_success(
        self,
        _,
        response,
        num_items,
        expected_value,
        extended_errors,
    ):
        succ, err = format_portfolios_response(
            response, extended_error_details=extended_errors
        )
        self.assert_responses(
            num_items, expected_value, succ=succ, err=err, err_extended=extended_errors
        )

        printer = self.create_printer(response, "portfolios", extended_errors)

        succ, err = printer.format_response()
        self.assert_responses(
            num_items, expected_value, succ=succ, err=err, err_extended=extended_errors
        )

    @parameterized.expand(
        [
            (
                "standard_response",
                responses,
                2,
                {"succ": ["code", "code"], "err": ["not found", "not found"]},
                False,
            ),
            ("empty_response", empty_response_with_full_shape, 0, {}, False),
            (
                "standard_response_with_extended_errors",
                responses,
                2,
                {"succ": ["code", "code"], "err": extended_error_expected},
                True,
            ),
        ]
    )
    def test_format_transactions_response_success(
        self, _, response, num_items, expected_value, extended_errors
    ):
        succ, err = format_transactions_response(
            response, extended_error_details=extended_errors
        )
        self.assert_responses(
            num_items, expected_value, succ=succ, err=err, err_extended=extended_errors
        )

        printer = self.create_printer(response, "transactions", extended_errors)

        succ, err = printer.format_response()
        self.assert_responses(
            num_items, expected_value, succ=succ, err=err, err_extended=extended_errors
        )

    @parameterized.expand(
        [
            (
                "standard_response",
                responses,
                2,
                {
                    "succ": ["BBG001MM1KV4", "BBG001MM1KV4"],
                    "failed": ["BBG001MM1KV4", "BBG001MM1KV4"],
                    "err": ["not found", "not found"],
                },
                False,
            ),
            ("empty_response", empty_response_with_full_shape, 0, {}, False),
            (
                "empty_response_missing_shape",
                empty_response_missing_shape,
                0,
                {},
                False,
            ),
            (
                "standard_response_with_extended_errors",
                responses,
                2,
                {
                    "succ": ["BBG001MM1KV4", "BBG001MM1KV4"],
                    "failed": ["BBG001MM1KV4", "BBG001MM1KV4"],
                    "err": extended_error_expected,
                },
                True,
            ),
        ]
    )
    def test_format_quotes_response_success(
        self, _, response, num_items, expected_value, extended_errors
    ):
        succ, err, failed = format_quotes_response(
            response, extended_error_details=extended_errors
        )
        self.assertEqual(num_items, len(succ))
        self.assertEqual(num_items, len(err))
        self.assertEqual(num_items, len(failed))
        quote_success_id_selector = "quoteId.quoteSeriesId.instrumentId"
        quote_failed_id_selector = "id"

        for index, row in succ.iterrows():
            self.assertEqual(
                expected_value["succ"][index],
                row[quote_success_id_selector],
            )
        self.assert_responses(
            num_items, expected_value, err=err, err_extended=extended_errors
        )
        for index, row in failed.iterrows():
            self.assertEqual(
                expected_value["failed"][index],
                row[quote_failed_id_selector],
            )

        printer = self.create_printer(response, "quotes", extended_errors)

        succ, err, failed = printer.format_response()
        self.assertEqual(num_items, len(succ))
        self.assertEqual(num_items, len(err))
        self.assertEqual(num_items, len(failed))

        for index, row in succ.iterrows():
            self.assertEqual(
                expected_value["succ"][index],
                row[quote_success_id_selector],
            )
        self.assert_responses(
            num_items, expected_value, err=err, err_extended=extended_errors
        )
        for index, row in failed.iterrows():
            self.assertEqual(
                expected_value["failed"][index],
                row[quote_failed_id_selector],
            )

    @parameterized.expand(
        [
            (
                "standard_response",
                responses,
                2,
                {"succ": ["code", "code"], "err": ["not found", "not found"]},
                False,
            ),
            ("empty_response", empty_response_with_full_shape, 0, {}, False),
            (
                "standard_response_with_extended_errors",
                responses,
                2,
                {"succ": ["code", "code"], "err": extended_error_expected},
                True,
            ),
        ]
    )
    def test_format_holdings_response_success(
        self, _, response, num_items, expected_value, extended_errors
    ):
        succ, err = format_holdings_response(
            response, extended_error_details=extended_errors
        )
        self.assert_responses(
            num_items, expected_value, succ=succ, err=err, err_extended=extended_errors
        )

        printer = self.create_printer(response, "holdings", extended_errors)

        succ, err = printer.format_response()
        self.assert_responses(
            num_items, expected_value, succ=succ, err=err, err_extended=extended_errors
        )

    @parameterized.expand(
        [
            (
                "standard_response",
                responses,
                2,
                {"succ": ["ID00002", "ID00002"], "err": ["not found", "not found"]},
                False,
            ),
            ("empty_response", empty_response_with_full_shape, 0, {}, False),
            (
                "standard_response_with_extended_errors",
                responses,
                2,
                {"succ": ["ID00002", "ID00002"], "err": extended_error_expected},
                True,
            ),
        ]
    )
    def test_format_reference_portfolios_response_success(
        self, _, response, num_items, expected_value, extended_errors
    ):
        succ, err = format_reference_portfolios_response(
            response, extended_error_details=extended_errors
        )
        self.assert_responses(
            num_items, expected_value, succ=succ, err=err, err_extended=extended_errors
        )

        printer = self.create_printer(response, "reference_portfolios", extended_errors)

        succ, err = printer.format_response()
        self.assert_responses(
            num_items, expected_value, succ=succ, err=err, err_extended=extended_errors
        )

    # Test failure cases

    @parameterized.expand(
        [
            ("no_error_field", responses_no_error_field, ValueError),
            ("no_success_field", responses_no_success_field, ValueError),
        ]
    )
    def test_format_instruments_response_fail(self, _, response, expected_error):
        with self.assertRaises(expected_error):
            succ, err, failed = format_instruments_response(response)

        with self.assertRaises(expected_error):
            printer = self.create_printer(response, "instruments")
            succ, err, failed = printer.format_response()

    @parameterized.expand(
        [
            ("no_error_field", responses_no_error_field, ValueError),
            ("no_success_field", responses_no_success_field, ValueError),
        ]
    )
    def test_format_portfolios_response_fail_no_error_field(
        self, _, response, expected_error
    ):
        with self.assertRaises(expected_error):
            succ, err, failed = format_portfolios_response(response)

        with self.assertRaises(expected_error):
            printer = self.create_printer(response, "portfolios")
            succ, err, failed = printer.format_response()

    @parameterized.expand(
        [
            ("no_error_field", responses_no_error_field, ValueError),
            ("no_success_field", responses_no_success_field, ValueError),
        ]
    )
    def test_format_transactions_response_fail_no_error_field(
        self, _, response, expected_error
    ):
        with self.assertRaises(expected_error):
            succ, err, failed = format_transactions_response(response)

        with self.assertRaises(expected_error):
            printer = self.create_printer(response, "transactions")
            succ, err, failed = printer.format_response()

    @parameterized.expand(
        [
            ("no_error_field", responses_no_error_field, ValueError),
            ("no_success_field", responses_no_success_field, ValueError),
        ]
    )
    def test_format_quotes_response_fail_no_error_field(
        self, _, response, expected_error
    ):
        with self.assertRaises(expected_error):
            succ, err, failed = format_quotes_response(response)

        with self.assertRaises(expected_error):
            printer = self.create_printer(response, "quotes")
            succ, err, failed = printer.format_response()

    @parameterized.expand(
        [
            ("no_error_field", responses_no_error_field, ValueError),
            ("no_success_field", responses_no_success_field, ValueError),
        ]
    )
    def test_format_holdings_response_fail_no_error_field(
        self, _, response, expected_error
    ):
        with self.assertRaises(expected_error):
            succ, err, failed = format_holdings_response(response)

        with self.assertRaises(expected_error):
            printer = self.create_printer(response, "holdings")
            succ, err, failed = printer.format_response()

    @parameterized.expand(
        [
            ("no_error_field", responses_no_error_field, ValueError),
            ("no_success_field", responses_no_success_field, ValueError),
        ]
    )
    def test_format_reference_portfolios_response_fail_no_error_field(
        self, _, response, expected_error
    ):
        with self.assertRaises(expected_error):
            succ, err, failed = format_reference_portfolios_response(response)

        with self.assertRaises(expected_error):
            printer = self.create_printer(response, "reference_portfolios")
            succ, err, failed = printer.format_response()

    def test_invalid_key(self):
        with self.assertRaises(ValueError) as context:
            CocoonPrinter({"foo": {}})

        self.assertTrue("Response contains invalid key" in str(context.exception))

    def test_no_keys(self):
        with self.assertRaises(ValueError) as context:
            CocoonPrinter({})

        self.assertTrue("Response doesn't contain any keys." in str(context.exception))

    def test_too_many_keys(self):
        with self.assertRaises(ValueError) as context:
            CocoonPrinter(
                {
                    "instruments": {},
                    "portfolios": {},
                }
            )

        self.assertEqual(
            "Response contains too many keys - only one is allowed, but received ['instruments', 'portfolios']",
            str(context.exception),
        )

    def test_instruments(self):
        printer = CocoonPrinter(
            {
                "instruments": {
                    "errors": [api_exception for _ in range(2)],
                    "success": [instrument_success for _ in range(2)],
                },
            }
        )

        succ, err, failed = printer.format_response()
        self.assert_responses(
            2,
            {
                "succ": [
                    instrument_id,
                    instrument_id,
                ],
                "failed": [
                    "error101",
                    "error101",
                ],
                "err": ["not found", "not found"],
            },
            succ=succ,
            err=err,
            failed=failed,
            err_extended=False,
        )

    def create_printer(
        self, response, entity_type, extended_errors=False, data_entity_details=False
    ):
        return CocoonPrinter(
            self.filter_response(response, entity_type),
            extended_error_details=extended_errors,
            data_entity_details=data_entity_details,
        )

    def filter_response(self, response, entity_type):
        return {k: v for k, v in response.items() if k == entity_type}



import unittest
from pathlib import Path
import lusid
import lusid.models as models

from finbourne_sdk_utils.cocoon.utilities import create_scope_id
from finbourne_sdk_utils.cocoon.transaction_type_upload import upsert_transaction_type_alias
import logging
import json
from copy import deepcopy

logger = logging.getLogger()
new_buy_transaction_type = "BUY-LPT-TEST"
new_sell_transaction_type = "SELL-LPT-TEST"
type_which_does_not_exist = create_scope_id()
txn_type_source = "LPT_TXN_CFG_TESTS"


class CocoonTestTransactionTypeUpload(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:

        secrets_file = Path(__file__).parent.parent.parent.joinpath("secrets.json")
        cls.api_factory = lusid.SyncApiClientFactory()
        
        cls.system_configuration_api = cls.api_factory.build(
            lusid.api.SystemConfigurationApi
        )

        cls.class_transaction_type_config = [
            models.TransactionConfigurationData(
                aliases=[
                    models.TransactionConfigurationTypeAlias(
                        type=new_buy_transaction_type,
                        description="LPT_TESTBUY1",
                        transaction_class="LPT_TESTBUY1",
                        transaction_group=txn_type_source,
                        source=txn_type_source,
                        transaction_roles="AllRoles",
                        is_default=False,
                    )
                ],
                movements=[
                    models.TransactionConfigurationMovementData(
                        movement_types="StockMovement",
                        movement_options=[],
                        side="Side1",
                        direction=1,
                        properties={},
                        mappings=[],
                    ),
                    models.TransactionConfigurationMovementData(
                        movement_types="CashCommitment",
                        movement_options=[],
                        side="Side2",
                        direction=1,
                        properties={},
                        mappings=[],
                    ),
                ],
                properties={},
            ),
            models.TransactionConfigurationData(
                aliases=[
                    models.TransactionConfigurationTypeAlias(
                        type=new_sell_transaction_type,
                        description="LPT_TESTSELL1",
                        transaction_class="LPT_TESTSELL1",
                        transaction_group=txn_type_source,
                        source=txn_type_source,
                        transaction_roles="AllRoles",
                        is_default=False,
                    )
                ],
                movements=[
                    models.TransactionConfigurationMovementData(
                        movement_types="StockMovement",
                        movement_options=[],
                        side="Side1",
                        direction=-1,
                        properties={},
                        mappings=[],
                    ),
                    models.TransactionConfigurationMovementData(
                        movement_types="CashCommitment",
                        movement_options=[],
                        side="Side2",
                        direction=-1,
                        properties={},
                        mappings=[],
                    ),
                ],
                properties={},
            ),
        ]

        for trans_type in cls.class_transaction_type_config:

            try:

                cls.create_transaction_response = cls.system_configuration_api.create_configuration_transaction_type(
                    transaction_configuration_data_request=trans_type
                )

            except lusid.exceptions.ApiException as e:
                if json.loads(e.body)["code"] == 231:
                    logger.info(json.loads(e.body)["title"])

    def test_update_current_alias_with_new_movements(self):

        new_movements_alias = deepcopy(self.class_transaction_type_config)[0]
        new_movements_alias.movements[0].direction = -1
        new_movements_alias.movements[1].direction = -1

        upsert_transaction_type_alias(self.api_factory, [new_movements_alias])

        get_transaction_types = (
            self.system_configuration_api.list_configuration_transaction_types()
        )

        uploaded_alias = []

        for trans_type in get_transaction_types.transaction_configs:
            for alias in trans_type.aliases:
                if alias.type == new_movements_alias.aliases[0].type and alias.source == txn_type_source:
                    uploaded_alias.append(trans_type)

        self.assertEqual(uploaded_alias[0], new_movements_alias)
        
    def test_update_alias_which_does_not_exist(self):

        alias_which_does_not_exist = deepcopy(self.class_transaction_type_config)[0]
        alias_which_does_not_exist.aliases[0].type = create_scope_id()

        upsert_transaction_type_alias(self.api_factory, [alias_which_does_not_exist])

        get_transaction_types = (
            self.system_configuration_api.list_configuration_transaction_types()
        )

        uploaded_alias = []

        for trans_type in get_transaction_types.transaction_configs:
            for alias in trans_type.aliases:
                if alias.type == alias_which_does_not_exist.aliases[0].type and alias.source == txn_type_source:
                    uploaded_alias.append(trans_type)

        self.assertEqual(uploaded_alias[0], alias_which_does_not_exist)


    def test_update_multiple_current_alias_with_new_movements(self):

        trans_type_with_multiple_alias = self.class_transaction_type_config.copy()

        trans_type_with_multiple_alias_1 = trans_type_with_multiple_alias[0]
        trans_type_with_multiple_alias_1.movements[0].movement_types = "CashAccrual"
        trans_type_with_multiple_alias_1.movements[1].movement_types = "CashAccrual"

        trans_type_with_multiple_alias_1 = trans_type_with_multiple_alias[1]
        trans_type_with_multiple_alias_1.movements[0].movement_types = "CashAccrual"
        trans_type_with_multiple_alias_1.movements[1].movement_types = "CashAccrual"

        upsert_transaction_type_alias(self.api_factory, trans_type_with_multiple_alias)

        get_transaction_types = (
            self.system_configuration_api.list_configuration_transaction_types()
        )

        uploaded_alias = []

        for trans_type in get_transaction_types.transaction_configs:
            for alias in trans_type.aliases:
                if alias.type in [new_buy_transaction_type, new_sell_transaction_type] and alias.source == txn_type_source:
                    uploaded_alias.append(trans_type)

        self.assertEqual(uploaded_alias, trans_type_with_multiple_alias)



import unittest
from pathlib import Path
import pandas as pd
import json
import lusid
import lusid.models as models

from finbourne_sdk_utils import cocoon as cocoon
from finbourne_sdk_utils.cocoon.utilities import create_scope_id
import datetime
from dateutil.tz import tzutc
import logging

logger = logging.getLogger()


class CocoonTestPortfolioGroup(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:

        cls.portfolio_scope = create_scope_id()
        
        cls.api_factory = lusid.SyncApiClientFactory()
        
        cls.unique_portfolios = pd.read_csv(
            Path(__file__).parent.joinpath(
                "data/port_group_tests/test_1_pg_create_with_portfolios.csv"
            )
        )["FundCode"].tolist()

        def create_portfolio_model(code):
            model = models.CreateTransactionPortfolioRequest(
                display_name=code,
                code=code,
                base_currency="GBP",
                description="Paper transaction portfolio",
                created="2020-02-25T00:00:00Z",
            )
            return model

        try:

            for code in cls.unique_portfolios:
                cls.api_factory.build(
                    lusid.api.TransactionPortfoliosApi
                ).create_portfolio(
                    scope=cls.portfolio_scope,
                    create_transaction_portfolio_request=create_portfolio_model(code),
                )
        except lusid.exceptions.ApiException as e:
            if e.status == 404:
                logger.error(f"The portfolio {code} already exists")

    def log_error_requests_title(cls, domain, responses):
        if len(responses.get(domain, {}).get("errors", [])) > 0:
            for error in responses[domain]["errors"]:
                return logger.error(json.loads(error.body)["title"])

    def csv_to_data_frame_with_scope(cls, csv, scope):
        data_frame = pd.read_csv(Path(__file__).parent.joinpath(csv))
        data_frame["Scope"] = scope
        return data_frame

    def cocoon_load_from_dataframe(
        cls,
        scope,
        data_frame,
        mapping_optional={"values.scope": "Scope", "values.code": "FundCode",},
        property_columns=[],
        properties_scope=None,
    ):

        return cocoon.cocoon.load_from_data_frame(
            api_factory=cls.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required={
                "code": "PortGroupCode",
                "display_name": "PortGroupDisplayName",
            },
            mapping_optional=mapping_optional,
            file_type="portfolio_group",
            property_columns=property_columns,
            properties_scope=properties_scope,
        )

    def test_01_pg_create_with_portfolios(self) -> None:

        """
        Test description:
        ------------------
        Here we test adding multiple new portfolio groups with multiple portfolios.

        Expected outcome:
        -----------------
        We expect one successful request/response per portfolio group with multiple portfolios.
        """

        test_case_scope = create_scope_id()
        data_frame = self.csv_to_data_frame_with_scope(
            "data/port_group_tests/test_1_pg_create_with_portfolios.csv",
            self.portfolio_scope,
        )

        responses = self.cocoon_load_from_dataframe(
            scope=test_case_scope, data_frame=data_frame
        )

        #self.log_error_requests_title("portfolio_groups", responses)

        # Test that there is a successful request per line in the dataframe
        self.assertEqual(
            len(
                [
                    port_group
                    for nested_group in [
                        port_group.portfolios
                        for port_group in responses["portfolio_groups"]["success"]
                    ]
                    for port_group in nested_group
                ]
            ),
            len(data_frame),
        )

        # Test that all the portfolios in the dataframe are in the request
        self.assertEqual(
            first=sorted(
                [
                    code.to_dict()
                    for code in responses["portfolio_groups"]["success"][0].portfolios
                ],
                key=lambda item: item.get("code"),
            ),
            second=sorted(
                [
                    lusid.models.ResourceId(
                        code=data_frame["FundCode"][1], scope=data_frame["Scope"][1]
                    ).to_dict(),
                    lusid.models.ResourceId(
                        code=data_frame["FundCode"][0], scope=data_frame["Scope"][0]
                    ).to_dict(),
                ],
                key=lambda item: item.get("code"),
            ),
        )

        self.assertEqual(
            first=responses["portfolio_groups"]["success"][1].portfolios,
            second=[
                lusid.models.ResourceId(
                    code=data_frame["FundCode"][2], scope=data_frame["Scope"][2]
                )
            ],
        )

        # Assert that by no unmatched_identifiers are returned in the response for portfolio_groups
        self.assertFalse(
            responses["portfolio_groups"].get("unmatched_identifiers", False)
        )

    def test_02_pg_create_with_no_portfolio(self) -> None:

        """
        Test description:
        -----------------
        Here we test adding one new portfolio group with no portfolios.

        Expected outcome:
        -----------------
        We expect one successful new portfolio group with no portfolios.

        """

        test_case_scope = create_scope_id()
        data_frame = self.csv_to_data_frame_with_scope(
            "data/port_group_tests/test_2_pg_create_with_no_portfolio.csv",
            self.portfolio_scope,
        )

        responses = self.cocoon_load_from_dataframe(
            scope=test_case_scope, data_frame=data_frame, mapping_optional={}
        )

        self.log_error_requests_title("portfolio_groups", responses)

        # Check that the portfolio group is created

        self.assertEqual(
            first=len(
                [
                    port_group.id
                    for port_group in responses["portfolio_groups"]["success"]
                ]
            ),
            second=len(data_frame),
        )

        # Check that the correct portfolio group code is used
        self.assertEqual(
            first=responses["portfolio_groups"]["success"][0].id,
            second=lusid.models.ResourceId(
                scope=test_case_scope, code=data_frame["PortGroupCode"].tolist()[0]
            ),
        )

        # Check that the portfolio group request has now portfolios
        self.assertTrue(
            len(responses["portfolio_groups"]["success"][0].portfolios) == 0
        )

    def test_03_pg_create_multiple_groups_no_portfolio(self) -> None:

        """
        Test description:
        -----------------
        Here we test adding multiple new portfolio group with no portfolios.

        Expected outcome
        -----------------
        We expect successful requests/responses for multiple new portfolio groups.
        """

        test_case_scope = create_scope_id()
        data_frame = self.csv_to_data_frame_with_scope(
            "data/port_group_tests/test_3_pg_create_multiple_groups_no_portfolio.csv",
            self.portfolio_scope,
        )

        responses = self.cocoon_load_from_dataframe(
            scope=test_case_scope, data_frame=data_frame, mapping_optional={}
        )

        self.log_error_requests_title("portfolio_groups", responses)

        # Check that there is a requet per portfolio group

        self.assertEqual(
            first=len(
                [
                    port_group.id
                    for port_group in responses["portfolio_groups"]["success"]
                ]
            ),
            second=len(data_frame),
        )

        # Check that the portfolio group code matches the code in the dataframe

        self.assertEqual(
            first=responses["portfolio_groups"]["success"][1].id,
            second=lusid.models.ResourceId(
                scope=test_case_scope, code=data_frame["PortGroupCode"].tolist()[1]
            ),
        )

    def test_04_pg_create_with_portfolio_not_exist(self):

        """
        Test description:
        -----------------
        Here we test attempting to add a portfolio which does not exist to a portfolio group.

        Expected outcome:
        -----------------
        We expect the entire request to fail with a PortfolioNotFound error.
        """

        test_case_scope = create_scope_id()
        data_frame = self.csv_to_data_frame_with_scope(
            "data/port_group_tests/test_4_pg_create_with_portfolio_not_exist.csv",
            self.portfolio_scope,
        )

        responses = self.cocoon_load_from_dataframe(
            scope=test_case_scope, data_frame=data_frame
        )

        self.log_error_requests_title("portfolio_groups", responses)

        # Check that LUSID cannot find the portfolio

        self.assertEqual(
            json.loads(responses["portfolio_groups"]["errors"][0].body)["name"],
            "PortfolioNotFound",
        )

        # Check there are no successful requests

        self.assertEqual(len(responses["portfolio_groups"]["success"]), 0)

    def test_05_pg_create_with_duplicate_portfolios(self):

        """
        Test description:
        -----------------
        Here we test attempting to add two of the same portfolios to a portfolio group.

        Expected result:
        ----------------
        We expect that each unique portfolio gets added and duplicates should be ignored.
        """

        test_case_scope = create_scope_id()
        data_frame = self.csv_to_data_frame_with_scope(
            "data/port_group_tests/test_5_pg_create_with_duplicate_portfolios.csv",
            self.portfolio_scope,
        )

        responses = self.cocoon_load_from_dataframe(
            scope=test_case_scope, data_frame=data_frame
        )

        self.log_error_requests_title("portfolio_groups", responses)

        data_frame.drop_duplicates(inplace=True)

        # Check that there is a request for each unique portfolio

        self.assertEqual(
            len(
                [
                    port_group
                    for nested_group in [
                        port_group.portfolios
                        for port_group in responses["portfolio_groups"]["success"]
                    ]
                    for port_group in nested_group
                ]
            ),
            len(data_frame),
        )

        # Check that a request is generated with unqiue portfolio only

        self.assertEqual(
            first=sorted(
                [
                    code.to_dict()
                    for code in responses["portfolio_groups"]["success"][0].portfolios
                ],
                key=lambda item: item.get("code"),
            ),
            second=sorted(
                [
                    lusid.models.ResourceId(
                        code=data_frame["FundCode"][0], scope=data_frame["Scope"][0]
                    ).to_dict(),
                    lusid.models.ResourceId(
                        code=data_frame["FundCode"][1], scope=data_frame["Scope"][1]
                    ).to_dict(),
                ],
                key=lambda item: item.get("code"),
            ),
        )

    def test_06_pg_create_duplicate_port_group(self):

        """
        Test description:
        -----------------
        Here we test create two of the same portfolio groups

        Expected results
        -----------------
        We expect one successful requesy for the portfolio group

        """

        test_case_scope = create_scope_id()
        data_frame = self.csv_to_data_frame_with_scope(
            "data/port_group_tests/test_6_pg_create_duplicate_port_group.csv",
            self.portfolio_scope,
        )

        responses = self.cocoon_load_from_dataframe(
            scope=test_case_scope, data_frame=data_frame, mapping_optional={}
        )

        self.log_error_requests_title("portfolio_groups", responses)

        # Check for one successful request

        self.assertEqual(len(responses["portfolio_groups"]["success"]), 1)

        # Check the successful request has same code as dataframe portfolio group

        self.assertEqual(
            first=responses["portfolio_groups"]["success"][0].id,
            second=lusid.models.ResourceId(
                scope=test_case_scope, code=data_frame["PortGroupCode"].tolist()[0]
            ),
        )

    def test_07_pg_create_with_properties(self) -> None:

        """
        Test description:
        -----------------
        Here we test creating a portfolio group with properties.

        Expected output:
        ----------------
        The response contains the upserted properties.
        """

        test_case_scope = create_scope_id()
        data_frame = self.csv_to_data_frame_with_scope(
            "data/port_group_tests/test_7_pg_create_with_properties.csv",
            self.portfolio_scope,
        )

        responses = self.cocoon_load_from_dataframe(
            scope=test_case_scope,
            data_frame=data_frame,
            property_columns=["location", "MifidFlag"],
            properties_scope=test_case_scope,
        )

        self.log_error_requests_title("portfolio_groups", responses)

        response_with_properties = self.api_factory.build(
            lusid.api.PortfolioGroupsApi
        ).get_group_properties(
            scope=test_case_scope, code=data_frame["PortGroupCode"].tolist()[0],
        )

        self.assertEqual(
            {
                "PortfolioGroup/"
                + test_case_scope
                + "/location": lusid.models.ModelProperty(
                    key="PortfolioGroup/" + test_case_scope + "/location",
                    value=lusid.models.PropertyValue(label_value="UK"),
                    effective_from=datetime.datetime.min.replace(tzinfo=tzutc()),
                    effective_until=datetime.datetime.max.replace(tzinfo=tzutc()),
                ),
                "PortfolioGroup/"
                + test_case_scope
                + "/MifidFlag": lusid.models.ModelProperty(
                    key="PortfolioGroup/" + test_case_scope + "/MifidFlag",
                    value=lusid.models.PropertyValue(label_value="Y"),
                    effective_from=datetime.datetime.min.replace(tzinfo=tzutc()),
                    effective_until=datetime.datetime.max.replace(tzinfo=tzutc()),
                ),
            },
            response_with_properties.properties,
        )

    def test_08_pg_add_bad_portfolio(self):

        """
        Description:
        ------------
        Here we test add a portfolio which does not exist to a current portfolio group.

        Expected results:
        -----------------
        The portfolio group is returned without the portfolios added.
        """

        test_case_scope = create_scope_id()
        data_frame = self.csv_to_data_frame_with_scope(
            "data/port_group_tests/test_8_pg_add_bad_portfolio.csv",
            self.portfolio_scope,
        )

        # Create the portfolio group as a seperate request
        port_group_request = lusid.models.CreatePortfolioGroupRequest(
            code=data_frame["PortGroupCode"][0],
            display_name=data_frame["PortGroupCode"][0],
        )

        self.api_factory.build(lusid.api.PortfolioGroupsApi).create_portfolio_group(
            scope=test_case_scope, create_portfolio_group_request=port_group_request
        )

        responses = self.cocoon_load_from_dataframe(
            scope=test_case_scope, data_frame=data_frame
        )

        self.log_error_requests_title("portfolio_groups", responses)

        self.assertTrue(len(responses["portfolio_groups"]["errors"]) == 0)

        self.assertEqual(
            first=responses["portfolio_groups"]["success"][0].id,
            second=lusid.models.ResourceId(
                scope=test_case_scope, code=data_frame["PortGroupCode"].tolist()[0]
            ),
        )

    def test_09_pg_add_duplicate_portfolio(self) -> None:

        """
        Description:
        ------------
        Here we test adding duplicate portfolios to a portfolio group.

        Expected outcome:
        -----------------
        We expect the one portfolio to be added.
        """

        test_case_scope = create_scope_id()
        data_frame = self.csv_to_data_frame_with_scope(
            "data/port_group_tests/test_9_pg_add_duplicate_portfolio.csv",
            self.portfolio_scope,
        )

        # Create the portfolio group as a seperate request
        port_group_request = lusid.models.CreatePortfolioGroupRequest(
            code=data_frame["PortGroupCode"][0],
            display_name=data_frame["PortGroupCode"][0],
        )

        self.api_factory.build(lusid.api.PortfolioGroupsApi).create_portfolio_group(
            scope=test_case_scope, create_portfolio_group_request=port_group_request
        )

        responses = self.cocoon_load_from_dataframe(
            scope=test_case_scope, data_frame=data_frame
        )

        self.log_error_requests_title("portfolio_groups", responses)

        self.assertEqual(
            first=responses["portfolio_groups"]["success"][0].portfolios[0],
            second=lusid.models.ResourceId(
                code=data_frame["FundCode"][0], scope=data_frame["Scope"][0]
            ),
        )

    def test_10_pg_add_no_new_portfolio(self) -> None:

        """
        Test description:
        ------------
        Here we test adding an existing portfolio to portfolio group.

        Expected result:
        ----------------
        The portfolio group response should be returned with one unmodified portfolio.
        """

        test_case_scope = create_scope_id()
        data_frame = self.csv_to_data_frame_with_scope(
            "data/port_group_tests/test_10_pg_add_no_new_portfolio.csv",
            self.portfolio_scope,
        )

        port_group_request = lusid.models.CreatePortfolioGroupRequest(
            code=data_frame["PortGroupCode"][0],
            display_name=data_frame["PortGroupCode"][0],
            values=[
                lusid.models.ResourceId(
                    code=data_frame["FundCode"][0], scope=self.portfolio_scope
                )
            ],
        )

        self.api_factory.build(lusid.api.PortfolioGroupsApi).create_portfolio_group(
            scope=test_case_scope, create_portfolio_group_request=port_group_request,
        )

        responses = self.cocoon_load_from_dataframe(
            scope=test_case_scope, data_frame=data_frame,
        )

        self.log_error_requests_title("portfolio_groups", responses)

        self.assertEqual(
            first=responses["portfolio_groups"]["success"][0].portfolios[0],
            second=lusid.models.ResourceId(
                code=data_frame["FundCode"][0], scope=data_frame["Scope"][0]
            ),
        )

    def test_11_pg_add_bad_and_good_portfolios(self):

        """
        Test description:
        -----------------
        Here we test updating a portfolio group with good and bad portfolios.

        Expected result:
        -----------------
        Good portfolios should be added and bad ones not added.

        """

        test_case_scope = create_scope_id()
        data_frame = self.csv_to_data_frame_with_scope(
            "data/port_group_tests/test_11_pg_add_bad_and_good_portfolios.csv",
            self.portfolio_scope,
        )

        # Create the portfolio group as a seperate request
        port_group_request = lusid.models.CreatePortfolioGroupRequest(
            code=data_frame["PortGroupCode"][0],
            display_name=data_frame["PortGroupCode"][0],
        )

        self.api_factory.build(lusid.api.PortfolioGroupsApi).create_portfolio_group(
            scope=test_case_scope, create_portfolio_group_request=port_group_request
        )

        responses = self.cocoon_load_from_dataframe(
            scope=test_case_scope, data_frame=data_frame
        )

        self.log_error_requests_title("portfolio_groups", responses)

        remove_dupe_df = data_frame[~data_frame["FundCode"].str.contains("BAD_PORT")]

        self.assertEqual(
            first=sorted(
                [
                    code.to_dict()
                    for code in responses["portfolio_groups"]["success"][0].portfolios
                ],
                key=lambda item: item.get("code"),
            ),
            second=sorted(
                [
                    lusid.models.ResourceId(
                        code=remove_dupe_df["FundCode"].tolist()[0],
                        scope=self.portfolio_scope,
                    ).to_dict(),
                    lusid.models.ResourceId(
                        code=remove_dupe_df["FundCode"].tolist()[1],
                        scope=self.portfolio_scope,
                    ).to_dict(),
                ],
                key=lambda item: item.get("code"),
            ),
        )

    def test_12_pg_add_portfolios_different_scopes(self) -> None:

        """
        Test description:
        -----------------
        Here we test adding portfolios with multiple scopes.

        Expected outcome:
        -----------------
        Request should be successful - returned with portfolios with multiple scopes.

        """

        test_case_scope = create_scope_id()
        data_frame = self.csv_to_data_frame_with_scope(
            "data/port_group_tests/test_12_pg_add_portfolios_different_scopes.csv",
            self.portfolio_scope,
        )

        port_scope_for_test = create_scope_id()
        self.api_factory.build(lusid.api.TransactionPortfoliosApi).create_portfolio(
            scope=port_scope_for_test,
            create_transaction_portfolio_request=models.CreateTransactionPortfolioRequest(
                display_name=data_frame["FundCode"][0],
                code=data_frame["FundCode"][0],
                base_currency="GBP",
            ),
        )

        port_group_request = lusid.models.CreatePortfolioGroupRequest(
            code=data_frame["PortGroupCode"][0],
            display_name=data_frame["PortGroupCode"][0],
            values=[
                lusid.models.ResourceId(
                    code=data_frame["FundCode"][0], scope=port_scope_for_test
                )
            ],
        )

        self.api_factory.build(lusid.api.PortfolioGroupsApi).create_portfolio_group(
            scope=test_case_scope, create_portfolio_group_request=port_group_request,
        )

        responses = self.cocoon_load_from_dataframe(
            scope=test_case_scope, data_frame=data_frame,
        )

        self.log_error_requests_title("portfolio_groups", responses)

        self.assertTrue(
            expr=all(
                [
                    id in responses["portfolio_groups"]["success"][0].portfolios
                    for id in [
                        lusid.models.ResourceId(
                            code=data_frame["FundCode"][1], scope=data_frame["Scope"][1]
                        ),
                        lusid.models.ResourceId(
                            code=data_frame["FundCode"][0], scope=port_scope_for_test
                        ),
                        lusid.models.ResourceId(
                            code=data_frame["FundCode"][0], scope=data_frame["Scope"][0]
                        ),
                        lusid.models.ResourceId(
                            code=data_frame["FundCode"][2], scope=data_frame["Scope"][2]
                        ),
                    ]
                ]
            )
        )



import os
import unittest
from pathlib import Path
import pandas as pd
import lusid

from finbourne_sdk_utils import cocoon as cocoon
from parameterized import parameterized
from finbourne_sdk_utils import logger
import logging


class CocoonTestsFailures(unittest.TestCase):
    api_factory = None

    @classmethod
    def setUpClass(cls) -> None:
        secrets_file = Path(__file__).parent.parent.parent.joinpath("secrets.json")
        cls.api_factory = lusid.SyncApiClientFactory()
        cls.logger = logger.LusidLogger(os.getenv("FBN_LOG_LEVEL", "info"))

    @parameterized.expand(
        [
            [
                "A file with a missing unique identifier for an instrument",
                "TestScope1",
                "data/global-fund-combined-instrument-master-missing-identifiers.csv",
                {"name": "instrument_name"},
                {},
                {"Figi": "figi", "Isin": "isin", "ClientInternal": "client_internal"},
                ["s&p rating", "moodys_rating", "currency"],
                "TestPropertiesScope1",
                "instruments",
                None,
                ValueError,
            ],
            [
                "No identifiers specified in the identifier mapping",
                "TestScope1",
                "data/global-fund-combined-instrument-master.csv",
                {"name": "instrument_name"},
                {},
                {},
                ["s&p rating", "moodys_rating", "currency"],
                "TestPropertiesScope1",
                "instruments",
                None,
                ValueError,
            ],
            [
                "Another way of missing the identifier mapping",
                "TestScope1",
                "data/global-fund-combined-instrument-master.csv",
                {"name": "instrument_name"},
                {},
                None,
                ["s&p rating", "moodys_rating", "currency"],
                "TestPropertiesScope1",
                "instruments",
                None,
                ValueError,
            ],
            [
                "A missing required attribute in the required mapping",
                "TestScope1",
                "data/global-fund-combined-instrument-master.csv",
                {},
                {},
                {"Figi": "figi", "Isin": "isin", "ClientInternal": "client_internal"},
                ["s&p rating", "moodys_rating", "currency"],
                "TestPropertiesScope1",
                "instruments",
                None,
                ValueError,
            ],
            [
                "None provided for the required mapping",
                "TestScope1",
                "data/global-fund-combined-instrument-master.csv",
                None,
                {},
                {"Figi": "figi", "Isin": "isin", "ClientInternal": "client_internal"},
                ["s&p rating", "moodys_rating", "currency"],
                "TestPropertiesScope1",
                "instruments",
                None,
                TypeError,
            ],
            [
                "A file with missing names for two instruments",
                "TestScope1",
                "data/global-fund-combined-instrument-master-missing-names.csv",
                {"name": "instrument_name"},
                {},
                {"Figi": "figi", "Isin": "isin", "ClientInternal": "client_internal"},
                ["s&p rating", "moodys_rating", "currency"],
                "TestPropertiesScope1",
                "instruments",
                None,
                ValueError,
            ],
            [
                "A property column list that contains a property not in the dataframe",
                "TestScope1",
                "data/global-fund-combined-instrument-master-missing-identifiers.csv",
                {"name": "instrument_name"},
                {},
                {"Figi": "figi", "Isin": "isin", "ClientInternal": "client_internal"},
                ["s&p rating", "moodys_rating", "currency", "transaction_currency"],
                "TestPropertiesScope1",
                "instruments",
                None,
                ValueError,
            ],
            [
                "Non unique portfolios in the file",
                "prime_broker_test",
                "data/metamorph_portfolios.csv",
                {
                    "code": "FundCode",
                    "display_name": "display_name",
                    "created": "created",
                    "base_currency": "base_currency",
                },
                {"description": "description", "accounting_method": None},
                {},
                ["base_currency"],
                "operations001",
                "portfolios",
                None,
                ValueError,
            ],
            [
                "Misnamed required parameter effective_date rather than effective_at",
                "prime_broker_test",
                "data/holdings-example-unique-date.csv",
                {
                    "code": "FundCode",
                    "effective_date": "Effective Date",
                    "tax_lots.units": "Quantity",
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                "operations001",
                "holdings",
                None,
                ValueError,
            ],
            [
                "Duplication of transaction ids",
                "prime_broker_test",
                "data/global-fund-combined-transactions-duplicate-id.csv",
                {
                    "code": "portfolio_code",
                    "transaction_id": "id",
                    "type": "transaction_type",
                    "transaction_date": "transaction_date",
                    "settlement_date": "settlement_date",
                    "units": "units",
                    "transaction_price.price": "transaction_price",
                    "transaction_price.type": "price_type",
                    "total_consideration.amount": "amount",
                    "total_consideration.currency": "trade_currency",
                },
                {"transaction_currency": "trade_currency"},
                {
                    "Isin": "isin",
                    "Figi": "figi",
                    "ClientInternal": "client_internal",
                    "Currency": "currency_transaction",
                },
                ["exposure_counterparty", "compls", "val", "location_region"],
                "operations001",
                "transaction",
                None,
                ValueError,
            ],
            [
                "Missing portfolio code",
                "prime_broker_test",
                "data/holdings-example-unique-date.csv",
                {"effective_at": "Effective Date", "tax_lots.units": "Quantity"},
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                "operations001",
                "holdings",
                None,
                ValueError,
            ],
            [
                "A file type that is not supported",
                "prime_broker_test",
                "data/holdings-example-unique-date.csv",
                {
                    "effective_at": "Effective Date",
                    "code": "Account Id",
                    "tax_lots.units": "Quantity",
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                "operations001",
                "reference",
                None,
                ValueError,
            ],
            [
                "Load with sub-holding-keys based on a column that does not exist in the file",
                "prime_broker_test",
                "data/holdings-example-unique-date.csv",
                {
                    "code": "FundCode",
                    "effective_at": "Effective Date",
                    "tax_lots.units": "Quantity",
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                "operations001",
                "holdings",
                ["Strategy Tag"],
                ValueError,
            ],
            [
                "Duplicate column in the source file with first column empty",
                "prime_broker_test",
                "data/holdings-example-unique-date-duplicate-column-empty.csv",
                {
                    "code": "FundCode",
                    "effective_at": "Effective Date",
                    "tax_lots.units": "Quantity",
                },
                {
                    "tax_lots.cost.amount": None,
                    "tax_lots.cost.currency": "Local Currency Code",
                    "tax_lots.portfolio_cost": None,
                    "tax_lots.price": None,
                    "tax_lots.purchase_date": None,
                    "tax_lots.settlement_date": None,
                },
                {
                    "Isin": "ISIN Security Identifier",
                    "Sedol": "SEDOL Security Identifier",
                    "Currency": "is_cash_with_currency",
                },
                ["Prime Broker"],
                "operations001",
                "holdings",
                None,
                ValueError,
            ],
        ]
    )
    def test_load_from_data_frame_failure(
        self,
        test_name,
        scope,
        file_name,
        mapping_required,
        mapping_optional,
        identifier_mapping,
        property_columns,
        properties_scope,
        file_type,
        sub_holding_keys,
        expected_exception,
    ) -> None:
        """
        Test for failure cases

        :param str scope: The scope of the portfolios to load the transactions into
        :param str file_name: The name of the test data file
        :param dict{str, str} mapping_required: The dictionary mapping the dataframe fields to LUSID's required base transaction/holding schema
        :param dict{str, str} mapping_optional: The dictionary mapping the dataframe fields to LUSID's optional base transaction/holding schema
        :param dict{str, str} identifier_mapping: The dictionary mapping of LUSID instrument identifiers to identifiers in the dataframe
        :param list[str] property_columns: The columns to create properties for
        :param str properties_scope: The scope to add the properties to
        :param str file_type: The file type to load
        :param any expected_exception: The expected exception

        :return: None
        """
        # ignore tests
        ignore = ["Non unique portfolios in the file", "Duplication of transaction ids"]

        if test_name in ignore:
            self.skipTest("not yet implemented")

        data_frame = pd.read_csv(Path(__file__).parent.joinpath(file_name))

        logging.info(f"Expecting exception {expected_exception}\n")
        
        with self.assertRaises(expected_exception):

            cocoon.cocoon.load_from_data_frame(
                api_factory=self.api_factory,
                scope=scope,
                data_frame=data_frame,
                mapping_required=mapping_required,
                mapping_optional=mapping_optional,
                file_type=file_type,
                identifier_mapping=identifier_mapping,
                property_columns=property_columns,
                properties_scope=properties_scope,
                sub_holding_keys=sub_holding_keys,
            )

    @parameterized.expand(
        [
            [
                "Multi-index DataFrame",
                "TestScope1",
                "data/global-fund-combined-instrument-master-missing-identifiers.csv",
                {"name": "instrument_name"},
                {},
                {"Figi": "figi", "Isin": "isin", "ClientInternal": "client_internal"},
                ["s&p rating", "moodys_rating", "currency"],
                "TestPropertiesScope1",
                "instruments",
                None,
                TypeError,
            ]
        ]
    )
    def test_load_from_data_frame_failure_multi_index_dataframe(
        self,
        _,
        scope,
        file_name,
        mapping_required,
        mapping_optional,
        identifier_mapping,
        property_columns,
        properties_scope,
        file_type,
        sub_holding_keys,
        expected_exception,
    ) -> None:
        """
        Test for failure cases

        :param str scope: The scope of the portfolios to load the transactions into
        :param str file_name: The name of the test data file
        :param dict{str, str} mapping_required: The dictionary mapping the dataframe fields to LUSID's required base transaction/holding schema
        :param dict{str, str} mapping_optional: The dictionary mapping the dataframe fields to LUSID's optional base transaction/holding schema
        :param dict{str, str} identifier_mapping: The dictionary mapping of LUSID instrument identifiers to identifiers in the dataframe
        :param list[str] property_columns: The columns to create properties for
        :param str properties_scope: The scope to add the properties to
        :param str file_type: The file type to load
        :param any expected_exception: The expected exception

        :return: None
        """

        data_frame = pd.read_csv(Path(__file__).parent.joinpath(file_name))
        columns = data_frame.columns

        # Make this a multi-index dataframe by grouping on the first two columns
        data_frame = data_frame.groupby([columns[0], columns[1]]).agg(
            {column: "first" for column in columns}
        )

        logging.info(f"Expecting exception {expected_exception}\n")   

        with self.assertRaises(expected_exception):

            cocoon.cocoon.load_from_data_frame(
                api_factory=self.api_factory,
                scope=scope,
                data_frame=data_frame,
                mapping_required=mapping_required,
                mapping_optional=mapping_optional,
                file_type=file_type,
                identifier_mapping=identifier_mapping,
                property_columns=property_columns,
                properties_scope=properties_scope,
                sub_holding_keys=sub_holding_keys,
            )



import datetime
import os
import unittest
import uuid
from pathlib import Path
import pandas as pd


from finbourne_sdk_utils import cocoon as cocoon
from parameterized import parameterized
import lusid
from finbourne_sdk_utils import logger
from finbourne_sdk_utils.cocoon.utilities import create_scope_id
from finbourne_sdk_utils.cocoon import load_from_data_frame

client_internal = "Instrument/default/ClientInternal"
sedol = "Instrument/default/Sedol"
name = "Instrument/default/Name"


class MockTransaction:
    def __init__(self, transaction_id, instrument_identifiers):
        self.transaction_id = transaction_id
        self.instrument_identifiers = instrument_identifiers


def transaction_and_instrument_identifiers(trd_0003=False, trd_0004=False):
    if trd_0003:
        trd_0003 = MockTransaction(
            transaction_id="trd_0003",
            instrument_identifiers={
                client_internal: "THIS_WILL_NOT_RESOLVE_1",
                sedol: "FAKESEDOL1",
                name: "THIS_WILL_NOT_RESOLVE_1",
            },
        )
    else:
        trd_0003 = None

    if trd_0004:
        trd_0004 = MockTransaction(
            transaction_id="trd_0004",
            instrument_identifiers={
                client_internal: "THIS_WILL_NOT_RESOLVE_2",
                sedol: "FAKESEDOL2",
                name: "THIS_WILL_NOT_RESOLVE_2",
            },
        )
    else:
        trd_0004 = None

    return [transaction for transaction in [trd_0003, trd_0004] if transaction]


def dict_for_comparison(list_of_transactions):
    return {
        transaction.transaction_id: transaction.instrument_identifiers
        for transaction in list_of_transactions
    }


class CocoonTestsTransactions(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:

        cls.api_factory = lusid.SyncApiClientFactory( )
        
        cls.logger = logger.LusidLogger(os.getenv("FBN_LOG_LEVEL", "info"))

    @parameterized.expand(
        [
            [
                "Test standard transaction load",
                "prime_broker_test",
                "data/global-fund-combined-transactions.csv",
                {
                    "code": "portfolio_code",
                    "transaction_id": "id",
                    "type": "transaction_type",
                    "transaction_date": "transaction_date",
                    "settlement_date": "settlement_date",
                    "units": "units",
                    "transaction_price.price": "transaction_price",
                    "transaction_price.type": "price_type",
                    "total_consideration.amount": "amount",
                    "total_consideration.currency": "trade_currency",
                },
                {"transaction_currency": "trade_currency"},
                {
                    "Isin": "isin",
                    "Figi": "figi",
                    "ClientInternal": "client_internal",
                    "Currency": "currency_transaction",
                },
                ["exposure_counterparty", "compls", "val", "location_region"],
                "operations001",
                None,
                lusid.models.Version,
            ],
            [
                "Add in some constants",
                "prime_broker_test",
                "data/global-fund-combined-transactions.csv",
                {
                    "code": "portfolio_code",
                    "transaction_id": "id",
                    "type": "transaction_type",
                    "transaction_date": "transaction_date",
                    "settlement_date": "settlement_date",
                    "units": "units",
                    "transaction_price.price": "transaction_price",
                    "transaction_price.type": "price_type",
                    "total_consideration.amount": "amount",
                    "total_consideration.currency": "trade_currency",
                },
                {"transaction_currency": "trade_currency", "source": "$default"},
                {
                    "Isin": "isin",
                    "Figi": "figi",
                    "ClientInternal": "client_internal",
                    "Currency": "currency_transaction",
                },
                ["exposure_counterparty", "compls", "val", "location_region"],
                "operations001",
                None,
                lusid.models.Version,
            ],
            [
                "Pass in None for some of properties_scope which accepts this",
                "prime_broker_test",
                "data/global-fund-combined-transactions.csv",
                {
                    "code": "portfolio_code",
                    "transaction_id": "id",
                    "type": "transaction_type",
                    "transaction_date": "transaction_date",
                    "settlement_date": "settlement_date",
                    "units": "units",
                    "transaction_price.price": "transaction_price",
                    "transaction_price.type": "price_type",
                    "total_consideration.amount": "amount",
                    "total_consideration.currency": "trade_currency",
                },
                {"transaction_currency": "trade_currency", "source": "$default"},
                {
                    "Isin": "isin",
                    "Figi": "figi",
                    "ClientInternal": "client_internal",
                    "Currency": "currency_transaction",
                },
                ["exposure_counterparty", "compls", "val", "location_region"],
                None,
                None,
                lusid.models.Version,
            ],
            [
                "Pass a constant for the portfolio code",
                "prime_broker_test",
                "data/global-fund-combined-transactions.csv",
                {
                    "code": "$GlobalCreditFund",
                    "transaction_id": "id",
                    "type": "transaction_type",
                    "transaction_date": "transaction_date",
                    "settlement_date": "settlement_date",
                    "units": "units",
                    "transaction_price.price": "transaction_price",
                    "transaction_price.type": "price_type",
                    "total_consideration.amount": "amount",
                    "total_consideration.currency": "trade_currency",
                },
                {"transaction_currency": "trade_currency", "source": "$default"},
                {
                    "Isin": "isin",
                    "Figi": "figi",
                    "ClientInternal": "client_internal",
                    "Currency": "currency_transaction",
                },
                ["exposure_counterparty", "compls", "val", "location_region"],
                None,
                None,
                lusid.models.Version,
            ],
            [
                "Try with a small batch",
                "prime_broker_test",
                "data/global-fund-combined-transactions.csv",
                {
                    "code": "$GlobalCreditFund",
                    "transaction_id": "id",
                    "type": "transaction_type",
                    "transaction_date": "transaction_date",
                    "settlement_date": "settlement_date",
                    "units": "units",
                    "transaction_price.price": "transaction_price",
                    "transaction_price.type": "price_type",
                    "total_consideration.amount": "amount",
                    "total_consideration.currency": "trade_currency",
                },
                {"transaction_currency": "trade_currency", "source": "$default"},
                {
                    "Isin": "isin",
                    "Figi": "figi",
                    "ClientInternal": "client_internal",
                    "Currency": "currency_transaction",
                },
                ["exposure_counterparty", "compls", "val", "location_region"],
                None,
                2,
                lusid.models.Version,
            ],
            [
                "Try with a small batch & random scope to ensure property cretaion",
                "prime_broker_test",
                "data/global-fund-combined-transactions.csv",
                {
                    "code": "$GlobalCreditFund",
                    "transaction_id": "id",
                    "type": "transaction_type",
                    "transaction_date": "transaction_date",
                    "settlement_date": "settlement_date",
                    "units": "units",
                    "transaction_price.price": "transaction_price",
                    "transaction_price.type": "price_type",
                    "total_consideration.amount": "amount",
                    "total_consideration.currency": "trade_currency",
                },
                {"transaction_currency": "trade_currency", "source": "$default"},
                {
                    "Isin": "isin",
                    "Figi": "figi",
                    "ClientInternal": "client_internal",
                    "Currency": "currency_transaction",
                },
                ["exposure_counterparty", "compls", "val", "location_region"],
                f"prime_broker_test_{create_scope_id()}",
                2,
                lusid.models.Version,
            ],
            [
                "Test standard transaction load with nulls for optional parameters",
                "prime_broker_test",
                "data/global-fund-combined-transactions-with-nulls.csv",
                {
                    "code": "portfolio_code",
                    "transaction_id": "id",
                    "type": "transaction_type",
                    "transaction_date": "transaction_date",
                    "settlement_date": "settlement_date",
                    "units": "units",
                    "transaction_price.price": "transaction_price",
                    "transaction_price.type": "price_type",
                    "total_consideration.amount": "amount",
                    "total_consideration.currency": "trade_currency",
                },
                {"transaction_currency": "trans_currency"},
                {
                    "Isin": "isin",
                    "Figi": "figi",
                    "ClientInternal": "client_internal",
                    "Currency": "currency_transaction",
                },
                ["exposure_counterparty", "compls", "val", "location_region"],
                "operations001",
                None,
                lusid.models.Version,
            ],
        ]
    )
    def test_load_from_data_frame_transactions_success(
            self,
            _,
            scope,
            file_name,
            mapping_required,
            mapping_optional,
            identifier_mapping,
            property_columns,
            properties_scope,
            batch_size,
            expected_outcome,
    ) -> None:
        """
        Test that transactions

        :param str scope: The scope of the portfolios to load the transactions into
        :param str file_name: The name of the test data file
        :param dict{str, str} mapping_required: The dictionary mapping the dataframe fields to LUSID's required base transaction/holding schema
        :param dict{str, str} mapping_optional: The dictionary mapping the dataframe fields to LUSID's optional base transaction/holding schema
        :param dict{str, str} identifier_mapping: The dictionary mapping of LUSID instrument identifiers to identifiers in the dataframe
        :param list[str] property_columns: The columns to create properties for
        :param str properties_scope: The scope to add the properties to
        :param any expected_outcome: The expected outcome

        :return: None
        """
        data_frame = pd.read_csv(Path(__file__).parent.joinpath(file_name))

        responses = cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required=mapping_required,
            mapping_optional=mapping_optional,
            file_type="transactions",
            identifier_mapping=identifier_mapping,
            property_columns=property_columns,
            properties_scope=properties_scope,
            batch_size=batch_size,
        )

        self.assertGreater(len(responses["transactions"]["success"]), 0)

        self.assertEqual(
            len(responses["transactions"]["errors"]),
            0,
            [
                (error.status, error.reason, error.body)
                for error in responses["transactions"]["errors"]
            ],
        )

        # Assert that by default no unmatched_identifiers are returned in the response
        self.assertFalse(responses["transactions"].get("unmatched_identifiers", False))

        self.assertTrue(
            expr=all(
                isinstance(success_response.version, expected_outcome)
                for success_response in responses["transactions"]["success"]
            )
        )

    @parameterized.expand(
        [
            [
                "Test standard transaction load",
                f"prime_broker_test_dict_{uuid.uuid4()}",
                {
                    "code": "portfolio_code",
                    "transaction_id": "id",
                    "type": "transaction_type",
                    "transaction_date": "transaction_date",
                    "settlement_date": "settlement_date",
                    "units": "units",
                    "transaction_price.price": "transaction_price",
                    "transaction_price.type": "price_type",
                    "total_consideration.amount": "amount",
                    "total_consideration.currency": "trade_currency",
                },
                {"transaction_currency": "trade_currency"},
                {
                    "Isin": "isin",
                    "Figi": "figi",
                    "ClientInternal": "client_internal",
                    "Currency": "currency_transaction",
                },
                ["location_region"],
                "operations001",
                "operations001",
                "location_region",
            ],
            [
                "Test standard transaction load with scope",
                f"prime_broker_test_dict_{uuid.uuid4()}",
                {
                    "code": "portfolio_code",
                    "transaction_id": "id",
                    "type": "transaction_type",
                    "transaction_date": "transaction_date",
                    "settlement_date": "settlement_date",
                    "units": "units",
                    "transaction_price.price": "transaction_price",
                    "transaction_price.type": "price_type",
                    "total_consideration.amount": "amount",
                    "total_consideration.currency": "trade_currency",
                },
                {"transaction_currency": "trade_currency"},
                {
                    "Isin": "isin",
                    "Figi": "figi",
                    "ClientInternal": "client_internal",
                    "Currency": "currency_transaction",
                },
                [
                    {
                        "scope": "foo",
                        "source": "location_region",
                        "target": "My_location",
                    }
                ],
                "operations001",
                "foo",
                "My_location",
            ],
        ]
    )
    def test_properties_dicts(
            self,
            _,
            scope,
            mapping_required,
            mapping_optional,
            identifier_mapping,
            property_columns,
            properties_scope,
            expected_property_scope,
            expected_property_code,
    ) -> None:
        """
        Test that transactions

        :param str scope: The scope of the portfolios to load the transactions into
        :param dict{str, str} mapping_required: The dictionary mapping the dataframe fields to LUSID's required base transaction/holding schema
        :param dict{str, str} mapping_optional: The dictionary mapping the dataframe fields to LUSID's optional base transaction/holding schema
        :param dict{str, str} identifier_mapping: The dictionary mapping of LUSID instrument identifiers to identifiers in the dataframe
        :param list[str] property_columns: The columns to create properties for
        :param str properties_scope: The scope to add the properties to
        :param str expected_property_scope: The expected scope
        :param str expected_property_code: The expected code

        :return: None
        """
        data_frame = pd.read_csv(
            Path(__file__).parent.joinpath("data/global-fund-combined-transactions.csv")
        )

        # Create the portfolio
        portfolio_response = cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required={
                "code": "portfolio_code",
                "display_name": "portfolio_code",
                "base_currency": "$GBP",
            },
            file_type="portfolio",
            mapping_optional={
                "created": f"${str(data_frame['transaction_date'].min())}"
            },
        )

        cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required=mapping_required,
            mapping_optional=mapping_optional,
            file_type="transactions",
            identifier_mapping=identifier_mapping,
            property_columns=property_columns,
            properties_scope=properties_scope,
            batch_size=None,
        )

        transactions_from_response = self.api_factory.build(
            lusid.api.TransactionPortfoliosApi
        ).get_transactions(
            scope=scope,
            code="GlobalCreditFund",
            property_keys=[
                f"Transaction/{expected_property_scope}/{expected_property_code}"
            ],
            from_transaction_date="2019-09-01T00:00Z",
            to_transaction_date="2019-09-12T00:00Z",
        )

        for tx in transactions_from_response.values:
            self.assertIsNotNone(
                tx.properties.get(
                    f"Transaction/{expected_property_scope}/{expected_property_code}"
                ),
                tx,
            )

        # Delete the portfolio at the end of the test
        for portfolio in portfolio_response.get("portfolios").get("success"):
            self.api_factory.build(lusid.api.PortfoliosApi).delete_portfolio(
                scope=scope, code=portfolio.id.code
            )
    @parameterized.expand(
        [
            [
                "Test standard transaction load",
                "data/global-fund-combined-transactions.csv",
                True,
                [],
            ],
            [
                "Test standard transaction load with two unresolved instruments",
                "data/global-fund-combined-transactions-unresolved-instruments.csv",
                True,
                [
                    "unresolved_tx01",
                    "unresolved_tx02",
                ],
            ],
        ]
    )
    def test_load_from_data_frame_transactions_success_with_correct_unmatched_identifiers(
            self,
            _,
            file_name,
            return_unmatched_items,
            expected_unmatched_transactions,
    ) -> None:
        """
        Test that transactions are uploaded and have the expected response from load_from_data_frame

        :param str file_name: The name of the test data file
        :param bool return_unmatched_items: A flag to request the return of all objects with unresolved instruments
        :param expected_unmatched_transactions: The expected unmatched transactions appended to the response

        :return: None
        """
        # Unchanged vars that have no need to be passed via param (they would count as duplicate lines)
        scope = "prime_broker_test"
        mapping_required = {
            "code": "portfolio_code",
            "transaction_id": "id",
            "type": "transaction_type",
            "transaction_date": "transaction_date",
            "settlement_date": "settlement_date",
            "units": "units",
            "transaction_price.price": "transaction_price",
            "transaction_price.type": "price_type",
            "total_consideration.amount": "amount",
            "total_consideration.currency": "trade_currency",
        }
        mapping_optional = {"transaction_currency": "trade_currency"}
        identifier_mapping = {
            "Isin": "isin",
            "Figi": "figi",
            "ClientInternal": "client_internal",
            "Currency": "currency_transaction",
        }
        property_columns = ["exposure_counterparty", "compls", "val", "location_region"]
        properties_scope = "operations001"
        batch_size = None
        expected_outcome = lusid.models.Version

        data_frame = pd.read_csv(Path(__file__).parent.joinpath(file_name))

        responses = cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required=mapping_required,
            mapping_optional=mapping_optional,
            file_type="transactions",
            identifier_mapping=identifier_mapping,
            property_columns=property_columns,
            properties_scope=properties_scope,
            batch_size=batch_size,
            return_unmatched_items=return_unmatched_items,
        )

        self.assertGreater(len(responses["transactions"]["success"]), 0)

        self.assertEqual(len(responses["transactions"]["errors"]), 0)

        # Assert that the unmatched_transactions returned are as expected for each case
        # First check the count of transactions
        self.assertEqual(
            len(responses["transactions"].get("unmatched_items", False)),
            len(expected_unmatched_transactions),
        )
        # Then check the transaction ids are the ones expected
        self.assertCountEqual(
            [
                transaction.transaction_id
                for transaction in responses["transactions"].get("unmatched_items", [])
            ],
            expected_unmatched_transactions,
        )

        self.assertTrue(
            expr=all(
                isinstance(success_response.version, expected_outcome)
                for success_response in responses["transactions"]["success"]
            )
        )

    def test_return_unmatched_transactions_extracts_relevant_transactions_and_instruments(
            self,
    ):
        scope = "unmatched_transactions_test"
        code = "MIS_INST_FUND"

        data_frame = pd.read_csv(
            Path(__file__).parent.joinpath(
                "data/transactions_unmatched_instruments.csv"
            )
        )

        # Create a test portfolio
        cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required={
                "code": "portfolio_code",
                "display_name": "portfolio_name",
                "base_currency": "base_currency",
            },
            file_type="portfolio",
            mapping_optional={"created": "$2000-01-01"},
        )

        # Upsert some transactions; one with known id, four with unknown ids
        cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required={
                "code": "portfolio_code",
                "transaction_id": "txn_id",
                "type": "txn_type",
                "transaction_date": "txn_trade_date",
                "settlement_date": "txn_settle_date",
                "units": "txn_units",
                "transaction_price.price": "txn_price",
                "transaction_price.type": "$Price",
                "total_consideration.amount": "txn_consideration",
                "total_consideration.currency": "currency",
            },
            file_type="transaction",
            identifier_mapping={
                "ClientInternal": "instrument_id",
                "Sedol": "sedol",
                "Ticker": "ticker",
                "Name": "name",
            },
            mapping_optional={},
        )

        # Call the return_unmatched_transactions method for dates that include two of the four unmatched
        response_focused = cocoon.cocoon.return_unmatched_transactions(
            self.api_factory,
            scope,
            code,
            from_transaction_date="2020-01-01",
            to_transaction_date="2020-01-03",
        )

        # Call the return_unmatched_transactions method for dates that include four of the four unmatched
        response_wide = cocoon.cocoon.return_unmatched_transactions(
            self.api_factory,
            scope,
            code,
            from_transaction_date="2000-01-01",
            to_transaction_date="2050-01-01",
        )
        self.assertEqual(len(response_wide), 4)

        # Assert that there are only two values returned
        self.assertEqual(len(response_focused), 2)
        # Assert that the transaction ids and instrument identifiers from the returned transactions match expectations
        response_dict = {
            transaction.transaction_id: transaction.instrument_identifiers
            for transaction in response_focused
        }
        self.assertEqual(
            response_dict.get("trd_0003"),
            {
                client_internal: "THIS_WILL_NOT_RESOLVE_1",
                sedol: "FAKESEDOL1",
                name: "THIS_WILL_NOT_RESOLVE_1",
            },
        )
        self.assertEqual(
            response_dict.get("trd_0004"),
            {
                client_internal: "THIS_WILL_NOT_RESOLVE_2",
                sedol: "FAKESEDOL2",
                name: "THIS_WILL_NOT_RESOLVE_2",
            },
        )

        # Delete the portfolio at the end of the test
        self.api_factory.build(lusid.api.PortfoliosApi).delete_portfolio(
            scope=scope, code=code
        )


    @parameterized.expand(
        [
            [
                "All returned unmatched transactions were in the upload",
                "data/transactions_unmatched_instruments_all_in_upload.csv",
                transaction_and_instrument_identifiers(trd_0003=True, trd_0004=True),
                transaction_and_instrument_identifiers(trd_0003=True, trd_0004=True),
            ],
            [
                "No returned unmatched transactions were in the upload",
                "data/transactions_unmatched_instruments_none_in_upload.csv",
                transaction_and_instrument_identifiers(trd_0003=True, trd_0004=True),
                transaction_and_instrument_identifiers(trd_0003=False, trd_0004=False),
            ],
            [
                "Some of the returned unmatched transactions were in the upload",
                "data/transactions_unmatched_instruments_some_in_upload.csv",
                transaction_and_instrument_identifiers(trd_0003=True, trd_0004=True),
                transaction_and_instrument_identifiers(trd_0003=True, trd_0004=False),
            ],
        ]
    )
    def test_filter_unmatched_transactions_method_only_returns_transactions_originally_present_in_dataframe(
            self,
            _,
            data_frame_path,
            unmatched_transactions,
            expected_filtered_unmatched_transactions,
    ):
        """
        Test that unmatched transactions that were not part of the current load_from_data_frame operation are
        filtered out of the final list.

        :param str _: The name of the test
        :param str data_frame_path: The name of the test data file
        :param list unmatched_transactions: The raw list of unmatched transactions to be filtered
        :param expected_filtered_unmatched_transactions: The expected unmatched transactions appended to the response

        :return: None
        """
        # Load the dataframe
        data_frame = pd.read_csv(Path(__file__).parent.joinpath(data_frame_path))

        # Filter the transactions to remove ones that were not part of the upload
        filtered_unmatched_transactions = cocoon.cocoon.filter_unmatched_transactions(
            data_frame=data_frame,
            mapping_required={"transaction_id": "txn_id"},
            unmatched_transactions=unmatched_transactions,
        )

        # Assert that the transaction ids and identifiers from the transactions returned match up to the expected ones
        self.assertCountEqual(
            dict_for_comparison(filtered_unmatched_transactions),
            dict_for_comparison(expected_filtered_unmatched_transactions),
        )


    def test_filter_unmatched_transactions_can_paginate_responses_for_2001_transactions_returned(
            self,
    ):
        """
        The GetTransactions API will only return up to 2,000 transactions per request. This test is to verify that
        LPT can handle responses of greater length, expected to be split across multiple pages.

        Returns
        -------
        Nothing.
        """
        scope = "unmatched_transactions_2k_test"
        identifier_mapping = {
            "Isin": "isin",
            "Figi": "figi",
            "ClientInternal": "client_internal",
            "Currency": "currency_transaction",
        }
        mapping_required = {
            "code": "portfolio_code",
            "transaction_id": "id",
            "type": "transaction_type",
            "transaction_date": "transaction_date",
            "settlement_date": "settlement_date",
            "units": "units",
            "transaction_price.price": "transaction_price",
            "transaction_price.type": "price_type",
            "total_consideration.amount": "amount",
            "total_consideration.currency": "trade_currency",
        }
        mapping_optional = {"transaction_currency": "trade_currency"}

        # Load the dataframe
        data_frame = pd.read_csv(
            Path(__file__).parent.joinpath(
                "data/global-fund-combined-transaction_2001.csv"
            )
        )

        # Create the portfolio
        portfolio_response = cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required={
                "code": "portfolio_code",
                "display_name": "portfolio_code",
                "base_currency": "$GBP",
            },
            file_type="portfolio",
            mapping_optional={"created": "created_date"},
        )

        # Assert that the portfolio was created without any issues
        self.assertEqual(len(portfolio_response.get("portfolios").get("errors")), 0)

        # Load in the transactions
        transactions_response = cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required=mapping_required,
            mapping_optional=mapping_optional,
            file_type="transactions",
            identifier_mapping=identifier_mapping,
            batch_size=None,
            return_unmatched_items=True,
        )

        self.assertGreater(len(transactions_response["transactions"]["success"]), 0)
        self.assertEqual(len(transactions_response["transactions"]["errors"]), 0)

        # Assert that the unmatched_items returned are as expected
        self.assertEqual(
            len(transactions_response["transactions"].get("unmatched_items", [])),
            2001,
        )

        # Delete the portfolio at the end of the test
        for portfolio in portfolio_response.get("portfolios").get("success"):
            self.api_factory.build(lusid.api.PortfoliosApi).delete_portfolio(
                scope=scope, code=portfolio.id.code
            )


    @parameterized.expand(
        [
            [
                "Test standard transaction load",
                "load_dataframe_test",
                "data/no-subholding-keys-transactions.csv",
                {
                    "code": "portfolio_code",
                    "transaction_id": "id",
                    "type": "transaction_type",
                    "transaction_date": "transaction_date",
                    "settlement_date": "settlement_date",
                    "units": "units",
                    "transaction_price.price": "transaction_price",
                    "transaction_price.type": "price_type",
                    "total_consideration.amount": "amount",
                    "total_consideration.currency": "trade_currency",
                },
                {"transaction_currency": "trade_currency"},
                {
                    "Isin": "isin",
                    "Figi": "figi",
                    "ClientInternal": "client_internal",
                    "Currency": "currency_transaction",
                },
                ["exposure_counterparty", "compls", "val", "location_region"],
                "load_dataframe_test",
                None,
                ["SHK_data"],
                ["Transaction/load_dataframe_test/SHK_data"],
            ]
        ]
    )
    def test_load_from_dataframe_non_existent_subholding_keys(
            self,
            _,
            scope,
            file_name,
            mapping_required,
            mapping_optional,
            identifier_mapping,
            property_columns,
            properties_scope,
            batch_size,
            sub_holding_keys,
            expected_sub_holdings_keys,
    ):
        """
        This checks whether load_from_data_frame creates subholding keys for transactions when they don't already exist.
        """

        # create the apis to reduce repeat definitions
        portfolios_api = self.api_factory.build(lusid.PortfoliosApi)
        transaction_portfolio_api = self.api_factory.build(
            lusid.TransactionPortfoliosApi
        )

        # make sure the portfolio doesn't already exist in scope
        portfolios = portfolios_api.list_portfolios_for_scope(scope).values

        for portfolio in portfolios:
            portfolios_api.delete_portfolio(scope, portfolio.id.code)

        # create the portfolio we're going to use
        transaction_portfolio_api.create_portfolio(
            scope,
            {
                "displayName": "test_load_from_dataframe_non_existent_subholding_keys portfolio",
                "code": "no-SHK",
                "baseCurrency": "GBP",
                "created": "2017-06-22T00:00:00.0000000+00:00",
            },
        )

        # load the data
        data_frame = pd.read_csv(Path(__file__).parent.joinpath(file_name))

        # load in the transactions along with the sub-holding keys
        cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required=mapping_required,
            mapping_optional=mapping_optional,
            file_type="transactions",
            identifier_mapping=identifier_mapping,
            property_columns=property_columns,
            properties_scope=properties_scope,
            batch_size=batch_size,
            sub_holding_keys=sub_holding_keys,
        )

        # we get the details from the portfolio
        portfolio_details = transaction_portfolio_api.get_details(
            scope=scope, code="no-SHK"
        )

        # get the transactions that we just loaded in
        transactions = transaction_portfolio_api.get_transactions(
            scope=scope, code="no-SHK"
        )

        # check that the sub holding key is a property of the two transactions
        self.assertTrue(
            "Transaction/load_dataframe_test/SHK_data"
            in transactions.values[0].properties.keys()
        )
        self.assertTrue(
            "Transaction/load_dataframe_test/SHK_data"
            in transactions.values[1].properties.keys()
        )

        # check that the property is a sub-holding key in the portfolio
        self.assertSetEqual(
            set(portfolio_details.sub_holding_keys), set(expected_sub_holdings_keys)
        )

    @parameterized.expand(
        [
            [
                "Test standard transaction load with no commit mode",
                "prime_broker_test",
                "data/global-fund-combined-transactions.csv",
                {
                    "code": "portfolio_code",
                    "transaction_id": "id",
                    "type": "transaction_type",
                    "transaction_date": "transaction_date",
                    "settlement_date": "settlement_date",
                    "units": "units",
                    "transaction_price.price": "transaction_price",
                    "transaction_price.type": "price_type",
                    "total_consideration.amount": "amount",
                    "total_consideration.currency": "trade_currency",
                },
                {"transaction_currency": "trade_currency"},
                {
                    "Isin": "isin",
                    "Figi": "figi",
                    "ClientInternal": "client_internal",
                    "Currency": "currency_transaction",
                },
                None,
                None
            ],
            [
                "Test standard transaction load with incorrect commit mode",
                "prime_broker_test",
                "data/global-fund-combined-transactions.csv",
                {
                    "code": "portfolio_code",
                    "transaction_id": "id",
                    "type": "transaction_type",
                    "transaction_date": "transaction_date",
                    "settlement_date": "settlement_date",
                    "units": "units",
                    "transaction_price.price": "transaction_price",
                    "transaction_price.type": "price_type",
                    "total_consideration.amount": "amount",
                    "total_consideration.currency": "trade_currency",
                },
                {"transaction_currency": "trade_currency"},
                {
                    "Isin": "isin",
                    "Figi": "figi",
                    "ClientInternal": "client_internal",
                    "Currency": "currency_transaction",
                },
                "Not_a_valid_commit_mode",
                None
            ],
            [
                "Test standard transaction load with partial commit mode",
                "prime_broker_test",
                "data/global-fund-combined-transactions.csv",
                {
                    "code": "portfolio_code",
                    "transaction_id": "id",
                    "type": "transaction_type",
                    "transaction_date": "transaction_date",
                    "settlement_date": "settlement_date",
                    "units": "units",
                    "transaction_price.price": "transaction_price",
                    "transaction_price.type": "price_type",
                    "total_consideration.amount": "amount",
                    "total_consideration.currency": "trade_currency",
                },
                {"transaction_currency": "trade_currency"},
                {
                    "Isin": "isin",
                    "Figi": "figi",
                    "ClientInternal": "client_internal",
                    "Currency": "currency_transaction",
                },
                "Partial",
                9
            ],
            [
                "Test standard transaction load with atomic commit mode",
                "prime_broker_test",
                "data/global-fund-combined-transactions.csv",
                {
                    "code": "portfolio_code",
                    "transaction_id": "id",
                    "type": "transaction_type",
                    "transaction_date": "transaction_date",
                    "settlement_date": "settlement_date",
                    "units": "units",
                    "transaction_price.price": "transaction_price",
                    "transaction_price.type": "price_type",
                    "total_consideration.amount": "amount",
                    "total_consideration.currency": "trade_currency",
                },
                {"transaction_currency": "trade_currency"},
                {
                    "Isin": "isin",
                    "Figi": "figi",
                    "ClientInternal": "client_internal",
                    "Currency": "currency_transaction",
                },
                "Atomic",
                9
            ],
        ]
    )
    def test_load_from_data_frame_transactions_with_commit_mode(
            self,
            _,
            scope,
            file_name,
            mapping_required,
            mapping_optional,
            identifier_mapping,
            transactions_commit_mode,
            expected_outcome
    ) -> None:
        """
        Test that transactions

        :param str scope: The scope of the portfolios to load the transactions into
        :param str file_name: The name of the test data file
        :param dict{str, str} mapping_required: The dictionary mapping the dataframe fields to LUSID's required base transaction/holding schema
        :param dict{str, str} mapping_optional: The dictionary mapping the dataframe fields to LUSID's optional base transaction/holding schema
        :param dict{str, str} identifier_mapping: The dictionary mapping of LUSID instrument identifiers to identifiers in the dataframe
        :param str transactions_commit_mode: The transaction commit mode to try and execute

        :return: None
        """
        # Arrange
        data_frame = pd.read_csv(Path(__file__).parent.joinpath(file_name))

        # Act
        if transactions_commit_mode is None or transactions_commit_mode not in ('Atomic', "Partial"):
            with self.assertRaises(KeyError):
                cocoon.cocoon.load_from_data_frame(
                    api_factory=self.api_factory,
                    scope=scope,
                    data_frame=data_frame,
                    mapping_required=mapping_required,
                    mapping_optional=mapping_optional,
                    file_type="transactions_with_commit_mode",
                    identifier_mapping=identifier_mapping
                )
            return True

        response = cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required=mapping_required,
            mapping_optional=mapping_optional,
            file_type="transactions_with_commit_mode",
            transactions_commit_mode=transactions_commit_mode,
            identifier_mapping=identifier_mapping
        )

        # Assert
        self.assertEqual(
            len(response["transactions_with_commit_modes"]["success"][0].values),
            expected_outcome
        )

    @parameterized.expand(
        [
            [
                "Test failed load with atomic commit mode",
                "prime_broker_test",
                "data/global-fund-combined-transactions-with-failed-transactions.csv",
                {
                    "code": "portfolio_code",
                    "transaction_id": "id",
                    "type": "transaction_type",
                    "transaction_date": "transaction_date",
                    "settlement_date": "settlement_date",
                    "units": "units",
                    "transaction_price.price": "transaction_price",
                    "transaction_price.type": "price_type",
                    "total_consideration.amount": "amount",
                    "total_consideration.currency": "trade_currency",
                },
                {"transaction_currency": "trade_currency"},
                {
                    "Isin": "isin",
                    "Figi": "figi",
                    "ClientInternal": "client_internal",
                    "Currency": "currency_transaction",
                },
                "Atomic",
                0
            ],

        ]
    )
    def test_load_from_data_frame_transactions_with_atomic_commit_mode_expect_failures(
            self,
            _,
            scope,
            file_name,
            mapping_required,
            mapping_optional,
            identifier_mapping,
            transactions_commit_mode,
            expected_outcome
    ) -> None:
        """
        Test that transactions

        :param str scope: The scope of the portfolios to load the transactions into
        :param str file_name: The name of the test data file
        :param dict{str, str} mapping_required: The dictionary mapping the dataframe fields to LUSID's required base transaction/holding schema
        :param dict{str, str} mapping_optional: The dictionary mapping the dataframe fields to LUSID's optional base transaction/holding schema
        :param dict{str, str} identifier_mapping: The dictionary mapping of LUSID instrument identifiers to identifiers in the dataframe
        :param str transactions_commit_mode: The transaction commit mode to try and execute

        :return: None
        """
        # Arrange
        data_frame = pd.read_csv(Path(__file__).parent.joinpath(file_name))

        # Act
        response = cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required=mapping_required,
            mapping_optional=mapping_optional,
            file_type="transactions_with_commit_mode",
            transactions_commit_mode=transactions_commit_mode,
            identifier_mapping=identifier_mapping
        )

        # Assert
        self.assertEqual(
            len(response["transactions_with_commit_modes"]["success"]),
            expected_outcome
        )

    @parameterized.expand(
        [
            [
                "Test failed load with partial commit mode",
                "prime_broker_test",
                "data/global-fund-combined-transactions-with-failed-transactions.csv",
                {
                    "code": "portfolio_code",
                    "transaction_id": "id",
                    "type": "transaction_type",
                    "transaction_date": "transaction_date",
                    "settlement_date": "settlement_date",
                    "units": "units",
                    "transaction_price.price": "transaction_price",
                    "transaction_price.type": "price_type",
                    "total_consideration.amount": "amount",
                    "total_consideration.currency": "trade_currency",
                },
                {"transaction_currency": "trade_currency"},
                {
                    "Isin": "isin",
                    "Figi": "figi",
                    "ClientInternal": "client_internal",
                    "Currency": "currency_transaction",
                },
                "Partial",
                8
            ]
        ]
    )
    def test_load_from_data_frame_transactions_with_partial_commit_mode_expect_failures(
            self,
            _,
            scope,
            file_name,
            mapping_required,
            mapping_optional,
            identifier_mapping,
            transactions_commit_mode,
            expected_outcome
    ) -> None:
        """
        Test that transactions

        :param str scope: The scope of the portfolios to load the transactions into
        :param str file_name: The name of the test data file
        :param dict{str, str} mapping_required: The dictionary mapping the dataframe fields to LUSID's required base transaction/holding schema
        :param dict{str, str} mapping_optional: The dictionary mapping the dataframe fields to LUSID's optional base transaction/holding schema
        :param dict{str, str} identifier_mapping: The dictionary mapping of LUSID instrument identifiers to identifiers in the dataframe
        :param str transactions_commit_mode: The transaction commit mode to try and execute

        :return: None
        """
        # Arrange
        data_frame = pd.read_csv(Path(__file__).parent.joinpath(file_name))

        # Act
        response = cocoon.cocoon.load_from_data_frame(
            api_factory=self.api_factory,
            scope=scope,
            data_frame=data_frame,
            mapping_required=mapping_required,
            mapping_optional=mapping_optional,
            file_type="transactions_with_commit_mode",
            transactions_commit_mode=transactions_commit_mode,
            identifier_mapping=identifier_mapping
        )

        # Assert
        self.assertEqual(
            len(response["transactions_with_commit_modes"]["success"][0].values),
            expected_outcome
        )

    @classmethod
    def tearDownClass(cls):
        # remove portfolios/properties created in test_load_from_dataframe_non_existent_subholding_keys
        try:
            cls.api_factory.build(
                lusid.PropertyDefinitionsApi
            ).delete_property_definition(
                "Transaction", "load_dataframe_test", "SHK_data"
            )
        except lusid.ApiException as e:
            if "domain" not in str(e.body) and "PropertyNotDefined" not in str(e.body):
                raise e

        try:
            cls.api_factory.build(lusid.PortfoliosApi).delete_portfolio(
                "load_dataframe_test", "no-SHK"
            )
        except lusid.ApiException as e:
            if "PortfolioNotFound" not in str(e.body):
                raise e



import datetime
from dateutil.tz import tzutc


def gen_transaction_data(txn_count, txn_date):
    return [
        {
            "counterpartyId": None,
            "instrumentIdentifiers": {
                "Instrument/default/LusidInstrumentId": "CCY_USD"
            },
            "settlementDate": txn_date,
            "totalConsideration": {"amount": 10, "currency": "USD"},
            "transactionCurrency": "USD",
            "transactionDate": txn_date,
            "transactionId": str(x) + "-test-transactions-" + str(txn_date),
            "transactionPrice": {"price": 1.0, "type": "Price"},
            "type": "FundsIn",
            "units": 10,
        }
        for x in range(txn_count)
    ]


group_portfolio_requests = {
    "test_flush_groups_sub_scope": [
        {
            "code": "testFlushSubGroup",
            "created": "2019-10-04T00:00:00.0000000+00:00",
            "values": [
                {
                    "scope": "test_flush_groups_sub_scope",
                    "code": "TestFlushPortfolio02",
                },
                {
                    "scope": "test_flush_groups_sub_scope",
                    "code": "TestFlushPortfolio03",
                },
            ],
            "displayName": "TestFlushSubGroup",
        },
        {
            "code": "testFlushGroupsClean",
            "created": "2019-10-04T00:00:00.0000000+00:00",
            "values": [
                {"scope": "test_flush_groups_sub_scope", "code": "TestFlushPortfolio01"}
            ],
            "subGroups": [
                {"scope": "test_flush_groups_sub_scope", "code": "testFlushSubGroup"}
            ],
            "displayName": "TestFlushGroupClean",
        },
    ],
    "test_flush_groups_scope_01": [
        {
            "code": "testFlushGroupsCleanNoSub",
            "created": "2019-10-04T00:00:00.0000000+00:00",
            "values": [
                {"scope": "test_flush_groups_scope_01", "code": "TestFlushPortfolio01"},
                {"scope": "test_flush_groups_scope_01", "code": "TestFlushPortfolio02"},
                {"scope": "test_flush_groups_scope_01", "code": "TestFlushPortfolio03"},
            ],
            "displayName": "TestFlushGroupCleanNoSub",
        },
        {
            "code": "testFlushGroupsMixedScopes",
            "created": "2019-10-04T00:00:00.0000000+00:00",
            "values": [
                {"scope": "test_flush_groups_scope_02", "code": "TestFlushPortfolio01"},
                {"scope": "test_flush_groups_scope_02", "code": "TestFlushPortfolio02"},
                {"scope": "test_flush_groups_scope_03", "code": "TestFlushPortfolio01"},
                {"scope": "test_flush_groups_scope_03", "code": "TestFlushPortfolio02"},
            ],
            "displayName": "TestFlushGroupMixed",
        },
        {
            "code": "testFlushGroupsSingle",
            "created": "2019-10-04T00:00:00.0000000+00:00",
            "values": [
                {"scope": "test_flush_groups_scope_02", "code": "TestFlushPortfolio03"}
            ],
            "displayName": "TestFlushGroupSingle",
        },
        {
            "code": "testFlushGroupsEmpty",
            "created": "2019-10-04T00:00:00.0000000+00:00",
            "values": [],
            "displayName": "TestFlushGroupEmpty",
        },
    ],
}



from finbourne_sdk_utils.cocoon.async_tools import run_in_executor, ThreadPool
import lusid
import asyncio
from functools import reduce
from typing import Dict, List
from finbourne_sdk_utils.cocoon.async_tools import (
    start_event_loop_new_thread,
    stop_event_loop_new_thread,
)


def _join_holdings(
    holdings_to_join: Dict[str, List[lusid.models.PortfolioHolding]],
    group_by_portfolio: bool = False,
    dict_key="GroupHoldings",
) -> Dict[str, List[lusid.models.PortfolioHolding]]:
    """
    This function joins the holdings together from multiple Portfolios into a single list of PortfolioHolding

    Currently the following constraints exist:

    1) Only support for instrument properties where the first one is taken
    2) Sub-holding-keys are ignored
    3) The portfolio currency is assumed to be the same across all Portfolios

    Parameters
    ----------
    holdings_to_join : Dict[str, List[lusid.models.PortfolioHolding]
        The dictionary of lists of PortfolioHolding keyed by the unique combination of Scope and Code
        for the Portfolio Group or Portfolios. These will either be returned as is or joined to form a new dictionary
        with a single key against a single list of PortfolioHolding.

    group_by_portfolio : bool
        Whether or not to group the holdings by portfolio

    dict_key : str
        The key to use for the merged holdings, usually the scope/code combination of the Portfolio Group

    Returns
    -------
    holdings_joined : Dict[str, List[lusid.models.PortfolioHolding]
        The joined dictionary of a list of PortfolioHolding which either has a single key for the Portfolio
        Group or a key for each Portfolio in the group
    """

    # If each portfolio should remain separated
    if group_by_portfolio:
        return holdings_to_join

    # Flatten the holdings into a single list
    all_holdings = [
        holding
        for holding_list in holdings_to_join.values()
        for holding in holding_list
    ]
    # Initialise a dictionary to hold the positions against a key
    all_holdings_keyed = {}
    # Construct the keyed dictionary using the LUSID instrument id, holding type and currency as the unique key for each holding
    [
        all_holdings_keyed.setdefault(
            f"{holding.instrument_uid}:{holding.holding_type}:{holding.cost.currency}",
            [],
        ).append(holding)
        for holding in all_holdings
    ]

    # Initialise a list to hold the joined holdings
    joined_holdings = []

    # Reduce the list of holdings against each key to a single holding
    for key, value in all_holdings_keyed.items():
        joined_holdings.append(
            lusid.models.PortfolioHolding(
                # Use the instrument_uid from the key
                instrument_uid=key.split(":")[0],
                # Use the holding type from the key
                holding_type=key.split(":")[1],
                # Add the metric fields together
                units=reduce((lambda x, y: x + y), list(map(lambda x: x.units, value))),
                settled_units=reduce(
                    (lambda x, y: x + y), list(map(lambda x: x.settled_units, value))
                ),
                cost=lusid.models.CurrencyAndAmount(
                    # Use the currency form the key
                    currency=key.split(":")[2],
                    amount=reduce(
                        (lambda x, y: x + y), list(map(lambda x: x.cost.amount, value))
                    ),
                ),
                cost_portfolio_ccy=lusid.models.CurrencyAndAmount(
                    # Use the first currency, these holdings could be from different portfolios, so the validity of this is questionable
                    currency=value[0].cost_portfolio_ccy.currency,
                    amount=reduce(
                        (lambda x, y: x + y),
                        list(map(lambda x: x.cost_portfolio_ccy.amount, value)),
                    ),
                ),
                # Takes the properties from the first value, only allows instrument properties
                properties={
                    property_key: value
                    for property_key, value in value[0].properties.items()
                    if value.key.split("/")[0] == "Instrument"
                },
            )
        )

    return {dict_key: joined_holdings}


@run_in_executor
def _get_portfolio_group(
    api_factory: lusid.SyncApiClientFactory, scope: str, code: str, **kwargs
) -> lusid.models.PortfolioGroup:
    """
    This function gets a Portfolio Group from LUSID.

    Parameters
    ----------
    api_factory : lusid.SyncApiClientFactory
        The api factory to use
    scope : str
        The scope of the Portfolio Group
    code : str
        The code of the Portfolio Group, with the scope this uniquely identifiers the Portfolio Group

    Returns
    -------
    response : lusid.models.PortfolioGroup
        The Portfolio Group

    Other Parameters
    ------
    effective_at : datetime
        The effective datetime at which to get the Portfolio Group
    as_at : datetime
        The as at datetime at which to get the Portfolio Group
    thread_pool
        The thread pool to run this function in
    """

    # Filter out the relevant keyword arguments as the LUSID API will raise an exception if given extras
    lusid_keyword_arguments = {
        key: value for key, value in kwargs.items() if key in ["effective_at", "as_at"]
    }

    # Call LUSID to get the portfolio group
    response = api_factory.build(lusid.api.PortfolioGroupsApi).get_portfolio_group(scope=scope, code=code, **lusid_keyword_arguments)

    return response


@run_in_executor
def _get_portfolio_holdings(
    api_factory: lusid.SyncApiClientFactory, scope: str, code: str, **kwargs
) -> Dict[str, List[lusid.models.PortfolioHolding]]:
    """
    This function gets the holdings of a Portfolio from LUSID.

    Parameters
    ----------
    api_factory : lusid.SyncApiClientFactory
        The api factory to use
    scope : str
        The scope of the Portfolio
    code : str
        The code of the Portfolio, with the scope this uniquely identifiers the Portfolio
    effective_at : datetime
        The effective datetime at which to get the Portfolio Group
    as_at : datetime
        The as at datetime at which to get the Portfolio Group
    filter : str
    by_taxlots : bool
    property_keys : list[str]

    Returns
    -------
    response : Dict[str, List[lusid.models.PortfolioHolding]]
        The list of PortfolioHolding keyed by the unique combination of the Portfolio's scope and code

    Other Parameters
    ------
    effective_at : datetime
        The effective datetime at which to get the Portfolio Group
    as_at : datetime
        The as at datetime at which to get the Portfolio Group
    filter : str
        The filter to use to filter the holdings
    by_taxlots : bool
        Whether or not to break the holdings down into individual tax lots
    property_keys : list[str]
        The list of property keys to decorate onto the holdings, must be from the Instrument domain
    thread_pool
        The thread pool to run this function in
    """

    # Filter out the relevant keyword arguments as the LUSID API will raise an exception if given extras
    lusid_keyword_arguments = {
        key: value
        for key, value in kwargs.items()
        if key in ["effective_at", "as_at", "filter", "by_taxlots", "property_keys"]
    }

    # Call LUSID to get the holdings for the Portfolio
    response = api_factory.build(lusid.api.TransactionPortfoliosApi).get_holdings(scope=scope, code=code, **lusid_keyword_arguments)

    # Key the response with the unique scope/code combination
    return {f"{scope} : {code}": response.values}


async def _get_holdings_for_group_recursive(
    api_factory: lusid.SyncApiClientFactory,
    group_scope: str,
    group_code: str,
    group_by_portfolio=False,
    **kwargs,
) -> Dict[str, List[lusid.models.PortfolioHolding]]:
    """
    This function recursively gets the holdings for a Portfolio Group in LUSID by making a request to get the holdings for
    each sub-group and portofolio and then joining the results together above the API.

    Parameters
    ----------
    api_factory : lusid.SyncApiClientFactory
        The api factory to use
    group_scope : str
        The scope of the Portfolio Group
    group_code : str
        The code of the Portfolio Group
    group_by_portfolio : bool
        Whether or not to group the holdings by Portfolio, if False will merge all Holdings together based on Instrument

    Returns
    -------
    Dict[str, List[lusid.models.PortfolioHolding]]
        The single set of holdings

    Other Parameters
    ------
    effective_at : datetime
        The effective datetime at which to get the Portfolio Group
    as_at : datetime
        The as at datetime at which to get the Portfolio Group
    filter : str
        The filter to use to filter the holdings
    by_taxlots : bool
        Whether or not to break the holdings down into individual tax lots
    property_keys : list[str]
        The list of property keys to decorate onto the holdings, must be from the Instrument domain
    thread_pool
        The thread pool to run this function in
    """

    # Get the details for the Portfolio Group including its sub-group and Portfolio members
    response = await _get_portfolio_group(
        api_factory, group_scope, group_code, **kwargs
    )
    portfolios = response.portfolios
    sub_groups = response.sub_groups

    # Get the holdings for each portfolio
    portfolio_holdings = await asyncio.gather(
        *[
            _get_portfolio_holdings(
                api_factory=api_factory,
                scope=portfolio.scope,
                code=portfolio.code,
                **kwargs,
            )
            for portfolio in portfolios
        ],
        return_exceptions=False,
    )

    # Turn the list of dictionaries into a single dictionary where the holdings are keyed by the Portfolio scope : code
    portfolio_holdings_keyed = {k: v for d in portfolio_holdings for k, v in d.items()}

    # Join the holdings across all portfolios together
    joined_portfolio_holdings = _join_holdings(
        portfolio_holdings_keyed,
        group_by_portfolio,
        dict_key=f"{group_scope} : {group_code}",
    )

    # If there aren't any sub-groups return these joined holdings
    if len(sub_groups) == 0:
        return joined_portfolio_holdings
    # Otherwise get the joined holdings for the sub-groups
    else:
        sub_group_holdings = await asyncio.gather(
            *[
                _get_holdings_for_group_recursive(
                    api_factory=api_factory,
                    group_scope=sub_group.scope,
                    group_code=sub_group.code,
                    group_by_portfolio=group_by_portfolio,
                    **kwargs,
                )
                for sub_group in sub_groups
            ],
            return_exceptions=False,
        )

        # Turn the list of dictionaries into a single dictionary where the holdings are keyed by the Portfolio scope : code
        sub_group_holdings_keyed = {
            k: v for d in sub_group_holdings for k, v in d.items()
        }

        # Join together the holdings across all the sub-groups
        joined_sub_group_holdings = _join_holdings(
            sub_group_holdings_keyed, group_by_portfolio, dict_key="SubGroupHoldings"
        )

        # There shouldn't be any overlap between the keys in these two sets
        if (
            len(
                set(joined_sub_group_holdings.keys()).intersection(
                    set(joined_portfolio_holdings)
                )
            )
            > 0
        ):
            raise ValueError(
                "There are duplicate Portfolios in the Portfolio Group. This is not currently supported. "
                "Please ensure that there are no duplicates in your Porfolio Group via sub-groups and try again"
            )

        # Join and return the joined sub-group holdings with the joined portfolio holdings
        return _join_holdings(
            {**joined_sub_group_holdings, **joined_portfolio_holdings},
            group_by_portfolio,
            dict_key=f"{group_scope} : {group_code}",
        )


def get_holdings_for_group(
    api_factory: lusid.SyncApiClientFactory,
    group_scope: str,
    group_code: str,
    group_by_portfolio: bool = False,
    num_threads=5,
    **kwargs,
) -> Dict[str, List[lusid.models.PortfolioHolding]]:
    """
    This function gets the holdings for a Portfolio Group in LUSID.

    Parameters
    ----------
    api_factory : lusid.SyncApiClientFactory
        The api factory to use
    group_scope : str
        The scope of the Portfolio Group
    group_code : str
        The code of the Portfolio Group
    group_by_portfolio : bool
        Whether or not to group the holdings by Portfolio, if False will merge all Holdings together based on Instrument
    num_threads : int
        The number of threads to use for asynchronous programming

    Returns
    -------
    group_holdings : Dict[str, List[lusid.models.PortfolioHolding]]
        The single set of holdings either keyed by the Portfolio scope/code or the Portfolio Group scope/code

    Other Parameters
    ------
    effective_at : datetime
        The effective datetime at which to get the Portfolio Group
    as_at : datetime
        The as at datetime at which to get the Portfolio Group
    filter : str
        The filter to use to filter the holdings
    by_taxlots : bool
        Whether or not to break the holdings down into individual tax lots
    property_keys : list[str]
        The list of property keys to decorate onto the holdings, must be from the Instrument domain
    """

    # Create a new thread pool to run the asynchronous tasks in
    thread_pool = ThreadPool(num_threads).thread_pool
    kwargs["thread_pool"] = thread_pool

    # Start a new event loop in a new thread, this is required to run inside a Jupyter notebook
    loop = start_event_loop_new_thread()

    # Get the responses from LUSID
    group_holdings = asyncio.run_coroutine_threadsafe(
        _get_holdings_for_group_recursive(
            api_factory=api_factory,
            group_scope=group_scope,
            group_code=group_code,
            group_by_portfolio=group_by_portfolio,
            **kwargs,
        ),
        loop,
    ).result()

    # Stop the additional event loop
    stop_event_loop_new_thread(loop)

    return group_holdings



import unittest
from finbourne_sdk_utils.jupyter_tools import toggle_code


class TestHideCodeButton(unittest.TestCase):
    def test_portfolio_stopper(self) -> None:

        test_toggle = toggle_code("Toggle Docstring")
        self.assertIsNone(test_toggle)



import os
import unittest
import finbourne_sdk_utils.cocoon as cocoon
from parameterized import parameterized
from finbourne_sdk_utils import logger


class CocoonUtilitiesTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.logger = logger.LusidLogger(os.getenv("FBN_LOG_LEVEL", "info"))

    @parameterized.expand(
        [
            ["Standard Base URL", "https://fbn-prd.lusid.com/api"],
            ["Base URL with forward slash suffix", "https://fbn-prd.lusid.com/api/"],
        ]
    )
    def test_get_swagger_dict_success(self, _, api_url):

        swagger_dict = cocoon.utilities.get_swagger_dict(api_url=api_url)

        self.assertTrue(expr=isinstance(swagger_dict, dict))

    def test_get_swagger_dict_fail(self):

        with self.assertRaises(ValueError):
            cocoon.utilities.get_swagger_dict(api_url="https://fbn-prd.lusid.com")



from finbourne_sdk_utils.lpt import lpt
from finbourne_sdk_utils.lpt import lse
from finbourne_sdk_utils.lpt import stdargs
from finbourne_sdk_utils.lpt import txn_config_yaml as tcy

TOOLNAME = "txn_cfg"
TOOLTIP = "Get/Set the transaction configuration"


def parse(extend=None, args=None):
    return (
        stdargs.Parser(
            "Get/Set transaction configuration", ["filename", "limit", "NODFQ", "asat"]
        )
        .add(
            "action",
            choices=("get", "set", "try"),
            help="get or set the config. 'Try' can be used to validate a custom encoding",
        )
        .add("--raw", action="store_true", help="use raw (non custom) encoding")
        .add("--json", action="store_true", help="display the json to be sent")
        .add("--group", action="store_true", help="set a single group")
        .add(
            "--force",
            action="store_true",
            help="set a single group, remove existing aliases for the group",
        )
        .extend(extend)
        .parse(args)
    )


def validate_group(txn_types, group):
    for txn in txn_types:
        for alias in txn.aliases:
            assert alias.transaction_group == group, "More than one group in the list"


def rem_groups(txn_types_old, group, arg):
    def still_valid(tt):
        for cand in tt.aliases:
            if cand.transaction_group != group:
                return True
        if arg is not True:
            raise AssertionError(
                "Existing group detected, use '--force' to remove them"
            )
        return False

    def clear_out_group(tt):
        check = len(tt.aliases)
        tt.aliases = [cand for cand in tt.aliases if cand.transaction_group != group]

        if len(tt.aliases) != check and arg != True:
            raise AssertionError(
                "Existing group detected, use '--force' to remove them"
            )
        return tt

    return [clear_out_group(t) for t in txn_types_old if still_valid(t)]


def merge_sets(txn_types_old, txn_types, arg):
    group = txn_types[0].aliases[0].transaction_group

    validate_group(txn_types, group)
    txn_types_clean = rem_groups(txn_types_old, group, arg)

    txn_types += txn_types_clean
    return txn_types


def process_args(api, args):
    y = tcy.TxnConfigYaml(api.models)

    if args.action == "get":

        def get_success(result):
            y.dump(
                y.TransactionSetConfigurationDataNoLinks(
                    result.content.transaction_configs, result.content.side_definitions
                ),
                args.filename,
                args.raw,
            )
            return None

        return api.call.list_configuration_transaction_types().bind(get_success)

    if args.action == "try":
        ffs = y.load(args.filename)
        y.dump(ffs, "{}-try".format(args.filename))

    if args.action == "set":

        def set_success(result):
            print(y.get_yaml(result.content))
            return None

        if args.group:
            txn_types = y.load(args.filename)

            result = api.call.list_configuration_transaction_types()

            if result.right is not None:
                txn_types_old = result.right.content
            else:
                raise ValueError("Api call did not return correct result")

            txn_types = y.load_update_str(
                y.get_yaml(merge_sets(txn_types_old, txn_types, args.force))
            )
        else:
            txn_types = y.load_update(args.filename)

        # y.dump(ffs,"{}-set".format(args.filename),True)
        if args.json:
            print(txn_types)
            return None
        else:
            return api.call.set_configuration_transaction_types(
                transaction_set_configuration_data_request=txn_types
            ).bind(set_success)


# Standalone tool
def main(parse=parse):
    lpt.standard_flow(parse, lse.connect, process_args)



import requests
from datetime import datetime
from datetime import timedelta
from collections import UserString

# Behaves like a string, but refreshes the
# OKTA credentials should they expire


class RefreshingToken(UserString):
    def __init__(self, token_url, token_request_body, headers):

        token_data = {"expires": datetime.now(), "credentials": ""}

        def get_token():
            if token_data["expires"] <= datetime.now():

                okta_response = requests.post(
                    token_url, data=token_request_body, headers=headers
                )
                if okta_response.status_code != 200:
                    print("OKTA authentication failed")
                    print(okta_response.text)
                    exit()

                d = dict(okta_response.json())
                token_data["expires"] = datetime.now() + timedelta(
                    seconds=d.get("expires_in", 3600) - 60
                )
                token_data["credentials"] = d["access_token"]

            return token_data["credentials"]

        self.token_fn = get_token

    # Call the token function to get the credentials
    # (refreshing them if necessary)
    # and then return the attribute for the resulting token
    def __getattribute__(self, name):
        token = object.__getattribute__(self, "token_fn")()
        if name == "data":
            return token
        return token.__getattribute__(name)
